var Camera = function(pixelWidth, pixelHeight)
{
   var _pixelWidth = pixelWidth;
   var _pixelHeight = pixelHeight;
   var _aspectRatio = pixelWidth / pixelHeight;

   var _position = new Vector();
   var _lookAt = new Vector();
   var _up = new Vector();

   var _transform = new Matrix();
   var _invTransform = _transform.inverse();

   var _nearClipPlane = 0.01;
   var _farClipPlane = 100;

   var _isOrthographic = true;
   // The orthographicsSize is half the size of the vertical viewing volume.
   // THe horizontal size of the viewing volume depends on the aspect ratio.
   var _orthographicSize = 1.0;
   // The field of view of camera in degrees. This is vertical field of view.
   var _fieldOfView = 60;

   var _projMatrix = new Matrix();

   this.setTransform = function(pos, lookAt, up)
   {
      _position = pos;
      _lookAt = lookAt;
      _up = up;

      _transform = Matrix.lookAt(
         pos.x, pos.y, pos.z,
         lookAt.x, lookAt.y, lookAt.z,
         up.x, up.y, up.z
      );

      _invTransform = _transform.inverse();
   };

   this.getTransform = function()
   {
      return _transform;
   };

   this.getPosition = function()
   {
      return new Vector(_invTransform.m[3], _invTransform.m[7], _invTransform.m[11]);
   }

   this.isOrthographic = function()
   {
      return _isOrthographic;
   };

   this.isPerspective = function()
   {
      return !_isOrthographic;
   };

   this.setClipPlane = function(near, far)
   {
      _nearClipPlane = near;
      _farClipPlane = far;
   };

   this.setOrthographic = function(orthoSize)
   {
      _isOrthographic = true;
      _orthographicSize = orthoSize;
   };

   this.setPerspective = function(fov)
   {
      _isOrthographic = false;
      _fieldOfView = fov;
   };

   this.getProjectionMatrix = function()
   {
      if(this.isOrthographic())
      {
         var vOrthoSize = _orthographicSize;
         var hOrthoSize = vOrthoSize * _aspectRatio;
         _projMatrix = Matrix.ortho(-hOrthoSize, hOrthoSize, -vOrthoSize, vOrthoSize, _nearClipPlane, _farClipPlane);
      }
      else
      {
         // persepctive camera
         _projMatrix = Matrix.perspective(_fieldOfView, _aspectRatio, _nearClipPlane, _farClipPlane);
      }

      return _projMatrix;
   };

   this.calculateViewSpaceFrustumCorners = function(z)
   {
      console.assert(this.isPerspective());

      var halfFov = 0.5 * _fieldOfView;
      var halfFovRad = halfFov * Math.PI / 180;
      var dy = z * Math.tan(halfFovRad);
      var dx = dy * _aspectRatio;


      /*
         d3     d2
           + - +
           |   |
           + - +
         d0     d1
       */

      return [
         new Vector(-dx, -dy, -z),
         new Vector( dx, -dy, -z),
         new Vector( dx,  dy, -z),
         new Vector(-dx,  dy, -z),
      ];

   };

};