"use strict";
// reference: photon-planes-min.js on https://benedikt-bitterli.me/photon-planes/webgl/
// Original copyright from Benedikt Bitterli's photon planes WebGL demo.
/*
   Photon planes WebGL demo
   Copyright (c) 2017 Benedikt Bitterli

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/

const WEBGL_DEBUG = false;

var TransientPhoton = function()
{
    var canvas = document.createElement('canvas');
    this.canvas = canvas;
    canvas.width = 640;
    canvas.height = 480;
    document.getElementById("content").prepend(canvas);

    this.width = canvas.width;
    this.height = canvas.height;

    this.setupGL();

    this.transientPhotonPlaneTest();
    //  NOTE: Code for start by pressing a button. Good for debug with Spector.js.
    //        Remember to unhide button in index.html.
    //  var test = this.transientPhotonPlaneTest.bind(this);
    //  const startElem = document.querySelector('button');
    //  startElem.addEventListener('click', test, {once: true});
};

TransientPhoton.prototype.setupGL = function()
{
    var gl = this.canvas.getContext('webgl2', {preserveDrawingBuffer: true});
    if(!gl)
    {
        throw new Error('WebGL not supported');
    }

    if (WEBGL_DEBUG)
    {
         function validateNoneOfTheArgsAreUndefined(functionName, args) {
            for (var ii = 0; ii < args.length; ++ii) {
               if (args[ii] === undefined) {
                  console.error("undefined passed to gl." + functionName + "(" +
                                 WebGLDebugUtils.glFunctionArgsToString(functionName, args) + ")");
               }
            }
         }

         gl = WebGLDebugUtils.makeDebugContext(gl, undefined, validateNoneOfTheArgsAreUndefined);
    }

    var floatExt = gl.getExtension("EXT_color_buffer_float");
    var floatLinExt = gl.getExtension("OES_texture_float_linear");
    var floatBufExt = gl.getExtension("WEBGL_color_buffer_float");
    if (!floatExt || !floatLinExt) throw new Error("Your platform does not support float textures");

    this.gl = gl;
    tgl.init(gl);
    RenderUtilities.init(tgl);
};

TransientPhoton.prototype.setupUI = function()
{
   var renderer = this.renderer;

   var cameraThetaSlider = new tui.Slider("camera-theta", 0, 360, true, true, function (theta) {
      this.setLabel("theta: " + theta);
      renderer.setCameraTheta(theta);
   });
   cameraThetaSlider.setValue(0);

   var timeCenterSlider = new tui.Slider("camera-time-center", 0, 1, true, false, function(t) {
      var center = renderer.setTimeCenter(t);
      this.setLabel("time center: " + Utilities.formatFloat(center));
   });
   timeCenterSlider.setValue(0.25 / 4.0);
   renderer.timeCenterSlider = timeCenterSlider;
   // var maxTimeCenter = renderer.params.maxTimeCenter;
   // timeCenterSlider.setValue(2.2 / maxTimeCenter);

   var timeWidthSlider = new tui.Slider("camera-time-width", 0.01, 1, true, false, function(dt) {
      renderer.setTimeWidth(dt);
      this.setLabel("width: " + Utilities.formatFloat(dt));
   });
   timeWidthSlider.setValue(0.1);
   // timeWidthSlider.setValue(0.04);

   var playControlSelector = new tui.ButtonGroup("play-control-selector", false, ['Play', 'Pause'], function (idx) {
      var isPlaying = idx == 0 ? true : false;
      renderer.setIsPlaying(isPlaying);
   });
   playControlSelector.select(1);

   var enableTransientSelector = new tui.ButtonGroup("enable-transient-selector", false, ['Enable', 'Disable'], function (idx) {
      var isTransientEnabled = idx == 0 ? true : false;
      renderer.setIsTransientEnabled(isTransientEnabled);

      var transientControls = document.getElementById('renderer-controls-transient');
      if(isTransientEnabled)
      {
         transientControls.style.display = '';
      }
      else
      {
         transientControls.style.display = 'none';
      }
   });
   enableTransientSelector.select(0);

   // setup primitive type selector
   this.isCameraWarped = false;
   this.primitive = "parallelepiped";

   var primitives = [];
   for(var prim in renderer.primitiveContexts)
   {
      if (!prim.includes("persp"))
         primitives.push(prim);
   }
   var primSelectCallback = function(idx)
   {
      var prim = primitives[idx];
      this.primitive = prim;
      renderer.setPrimitive(prim, this.isCameraWarped);
   }.bind(this);

   var isCameraWarpedSelector = new tui.ButtonGroup("camera-warped-selector", false, ['Disable', 'Enable'], function (idx) {
      var isEnabled = idx == 0 ? false : true;
      this.isCameraWarped = isEnabled;
      primSelectCallback(primitives.indexOf(this.primitive), this.isCameraWarped);
   }.bind(this));
   isCameraWarpedSelector.select(0);

   var primitiveSelector = new tui.ButtonGroup("primitive-selector", false, primitives, primSelectCallback);
   primitiveSelector.select(primitives.indexOf(this.primitive));

   // Debug sliders
   var photonIdSlider = new tui.Slider("photon-id", -1, renderer.params.maxPerFrameVertexCount, true, true, function(photonId){
      if(photonId == -1)
      {
         renderer.setDrawSinglePhoton(false);
         this.setLabel("Disabled");
      }
      else
      {
         renderer.setDrawSinglePhoton(true, photonId);
         this.setLabel("Rendering photon: " + photonId);
      }
   });
   photonIdSlider.setValue(-1);
   photonIdSlider.parent.style.display = 'none';

   // exposure slider
   const stepsPerEv = 3;
   const maxEv = 3;
   var exposureSlider = new tui.Slider("camera-exposure", -stepsPerEv * maxEv, stepsPerEv * maxEv, true, true, function(evSteps) {
      let ev = evSteps / stepsPerEv;
      let exposure = Math.pow(2, ev);
      renderer.setExposure(exposure);

      let sign = Math.sign(ev);
      let evIntPart = Math.sign(ev) * Math.floor(Math.abs(ev));
      let evDeciPartNumerator = Math.abs(evSteps) % stepsPerEv;

      let label = "EV: ";
      if(evSteps < 0)
      {
         label += "-"
      }
      if(!(evIntPart == 0 && evDeciPartNumerator != 0))
      {
         label += Math.abs(evIntPart);
      }
      if(evDeciPartNumerator != 0)
      {
         if(evIntPart != 0)
         {
            label += " ";
         }
         label += evDeciPartNumerator + "/" + stepsPerEv;
      }
      this.setLabel(label);
   });

   // set ev compensate to 0
   exposureSlider.setValue(0 * stepsPerEv);

   // sample count (max # of paths) slider
   var targetPathCountSlider = new tui.ExpSlider("target-path-count", 1E2, 3E7, true, true, function(targetPathCount){
      this.setLabel("target path count: " + targetPathCount);

      renderer.setTotalPathCount(targetPathCount);
   });
   targetPathCountSlider.setValue(5e4);
   // targetPathCountSlider.setValue(3e5);

   // progress bar
   this.progressBar = new tui.ProgressBar("render-progress", true);

   // save image button
   var saveImageButton = document.getElementById('save-image-button');
   var saveImageFunc = renderer.saveAsImage.bind(renderer);
   saveImageButton.addEventListener('click', saveImageFunc, {once: false});

   // save as png
   var savePngButton = document.getElementById('save-image-png-button');
   var savePngFunc = renderer.saveAsPngImage.bind(renderer);
   savePngButton.addEventListener('click', savePngFunc, {once: false});
};

TransientPhoton.prototype.drawTriangleTest = function()
{
    var gl = this.gl;
    var shader = new tgl.Shader("simple-vert.shader", "simple-frag.shader");

    var vertexBuffer = new tgl.VertexBuffer();
    vertexBuffer.addAttribute('a_Position', gl.FLOAT, 2);
    vertexBuffer.init(3);

    vertexBuffer.copyAttribute('a_Position', new Float32Array([
        0, 0,
        0, 0.5,
        0.7, 0,
    ]));

    shader.bind();
    vertexBuffer.draw(shader, gl.TRIANGLES);

};

TransientPhoton.prototype.transientPhotonPlaneTest = function()
{
   this.renderer = new PhotonRenderer(this.gl, this.width, this.height);
   this.setupUI();

   this.renderer.render(0);
   this.boundRenderLoop = this.renderLoop.bind(this);

   window.requestAnimationFrame(this.boundRenderLoop);
};

TransientPhoton.prototype.renderLoop = function(timestamp)
{
   window.requestAnimationFrame(this.boundRenderLoop);
   this.renderer.update(timestamp);
   this.renderer.render(timestamp);

   this.progressBar.setProgress(this.renderer.getProgress());

   let total = this.renderer.getTotalPathCount();
   let traced = this.renderer.getTracedPathCount();
   let pathsPerSecond = Math.round(this.renderer.getPathPerSecond());
   let secondsElapsed = this.renderer.getSecondsElapsed();
   let secondsPart = Math.floor(secondsElapsed);
   let millisecondsPart = Math.round((secondsElapsed - secondsPart) * 1000.0);
   let timeStr = "";
   if (secondsPart > 0)  timeStr += secondsPart + "s ";
   timeStr += millisecondsPart + "ms";

   this.progressBar.setLabel(traced + "/" + total + " paths traced; Progress: " + this.progressBar.getProgressPercentage() + "%" + "; Paths per second: " + pathsPerSecond + "; Time elapsed: " + timeStr);
};

//// PhotonRendererParams: stores the data/objects shared between PhotonRenderer (parent renderer)
////                       and PrimitiveRenderer (child renderer)
var PhotonRendererParams = function(gl, width, height)
{
   //// Env Params:
   this.sigmaS = new Vector(1, 1, 1);
   this.sigmaA = new Vector(1, 1, 1);
   this.sigmaT = this.sigmaS.add(this.sigmaA);

   //// Photon Params:
   this.numLightBounces = 3;
   this.totalPathCount = 1E5;
   this.initPerFrameVertexCount = 5E3;
   this.perFrameVertexCount = this.initPerFrameVertexCount;
   this.maxPerFrameVertexCount = 1E7;
   this.minPerFrameVertexCount = 1E1;

   this.tracedPhotonCount = 0;
   this.tracedPathCount = 0;

   this.lastTracedPhotonCount = 0;
   this.lastTracedPathCount = 0;

   this.phaseG = 0;
   this.tracer = new SearchLightPhotonTracer(this.sigmaS, this.sigmaA, this.phaseG, this.maxPerFrameVertexCount);
   this.rng = new Xorshift32;

   this.evalCamPhase = true;
   this.evalCamTransmittance = true;

   //// Transient Params:
   // photons (ray paths) in the time interval
   // (timeCenter - timeWidth / 2., timeCenter + timeWidth / 2.)
   // will be counted when transient rendering is enabled
   this.isTransientEnabled = true;
   this.timeCenter = 1;
   this.timeWidth = 0.01;
   this.speedOfLight = 1;
   this.isTimeWidthZero = false;

   //// Playing Params:
   this.maxTimeCenter = 4;
   this.isPlaying = false;
   this.realTimeSinceLastPlayedFrame = 0.0;
   // seconds to reach the max time center from 0, if isPlaying is enabled
   this.reachMaxTimeSeconds = 2.0;
   // timestamp (also for seeding the rand() in shader)
   this.timestamp = 0;

   //// GL/Canvas Params:
   this.gl = gl;
   this.width = width;
   this.height = height;
   this.aspectRatio = this.width / this.height;
   this.camera = new Camera(width, height);

   this.perspCamera = new Camera(width, height);
   this.perspCamera.setPerspective(90);

   //// Debug Mode Params:
   this.singlePhotonMode = {
      "isEnabled"  : false,
      "photonId" : -1,
   };
};

PhotonRendererParams.prototype.getLightPower = function()
{
   if (!this.evalCamTransmittance)
   {
      return 10;
   }
   else
   {
      return 1e3;
   }
};



//// PhotonRenderer: class for rendering photons
var PhotonRenderer = function(gl, width, height)
{
   this.prevTimestamp = 0;
   this.lastResetTimestamp = 0;
   this.lastFinishedTimestamp = 0;

   this.gl = gl;
   this.params = new PhotonRendererParams(gl, width, height);

   // time width zero slider
   var isTimeWidthZeroSelector = new tui.ButtonGroup("zero-time-width-selector", false, ['Disable', 'Enable'], function (idx) {
      var isEnabled = idx == 0 ? false : true;
      this.params.isTimeWidthZero = isEnabled;
      this.reset();
   }.bind(this));
   isTimeWidthZeroSelector.select(0);
   this.isTimeWidthZeroSelector = isTimeWidthZeroSelector;
   this.setIsTimeWidthZeroSelectorEnabled(false);

   // # of light bounces slider
   var params = this.params;
   var photonRenderer = this;
   var numLightBouncesSlider = new tui.Slider("num-light-bounces", 1, 3, true, true, function(numLightBounces){
      this.setLabel("# light bounces: " + numLightBounces);
      params.numLightBounces = numLightBounces;
      photonRenderer.reset();
   });
   numLightBouncesSlider.setValue(3);

   this.camera = this.params.camera;
   // set orthographic camera parameters
   this.camera.setOrthographic(1);

   // set perspective camera parameters
   // this.camera.setPerspective(90.0);

   // for dynamically adjusting # of vertices (photons) to draw per frame
   this.deltaTimes = [];

   // enable photon tracer
   this.tracePhotons();

   this.quadVbo = RenderUtilities.getQuadVbo();
   this.blitShader = new tgl.Shader("blit-vert.shader", "blit-frag.shader", {'ENABLE_GAMMA_CORRECTION': 1});
   this.meanShader = new tgl.Shader("blit-vert.shader", "blit-running-mean-frag.shader");

   this.renderSingleContext = false; // only render the current selected prim instead of all, good for profiling

   this.primitiveContexts = {
      "beam"  : {
         "renderer" : new BeamRenderer(this),
      },
      "plane" : {
         "renderer" : new PlaneRenderer(this),
      },
      "parallelepiped": {
         "renderer" : new VolumeRenderer(this),
      },
      "ball": {
         "renderer" : new SphereRenderer(this),
      },
      "beam_persp" : {
         "renderer" : new BeamRenderer(this, true),
      },
      "plane_persp" : {
         "renderer" : new PlaneRenderer(this, true),
      },
      "parallelepiped_persp" : {
         "renderer" : new VolumeRenderer(this, true),
      },
      "ball_persp" : {
         "renderer" : new EllipsoidRenderer(this),
      },
   };

   // create a framebuffer and an fbo for each context
   for(var key in this.primitiveContexts)
   {
      var context = this.primitiveContexts[key];
      context.buffer = new tgl.Texture(this.params.width, this.params.height, 4, true, false, true);
      context.fbo = new tgl.RenderTarget(context.buffer);

      context.frontBufferIdx = 0;
      context.avgBuffers = [null, null];
      context.avgFbos = [null, null];
      for(var i = 0; i < context.avgBuffers.length; ++i)
      {
         context.avgBuffers[i] = new tgl.Texture(this.params.width, this.params.height, 4, true, false, true);
         context.avgFbos[i] = new tgl.RenderTarget(context.avgBuffers[i]);
         context.avgFbos[i].clearColor(0, 0, 0, 0);
      }
   }

   this.activeContext = this.primitiveContexts["plane"];
   this.setIsTransientEnabled(this.params.isTransientEnabled);

   this.exposure = 1.0;

   this.imageFile = null;
};

PhotonRenderer.prototype.tracePhotons = function()
{
   var sampleInfo = this.params.tracer.generateSamples(
      this.params.getLightPower(),
      3,
      this.params.perFrameVertexCount,
      this.params.totalPathCount - this.params.tracedPathCount,
      this.params.numLightBounces == 1 ? false : true
   );

   var pathCount = sampleInfo[0];
   var photonCount = sampleInfo[1];

   this.params.lastTracedPathCount = pathCount;
   this.params.lastTracedPhotonCount = photonCount;

   this.params.tracedPathCount += pathCount;
   this.params.tracedPhotonCount += photonCount;

   console.assert(this.params.tracedPathCount <= this.params.totalPathCount);
};

// update() is called before render.
PhotonRenderer.prototype.update = function(timestamp)
{
   this.params.timestamp = timestamp;
   var deltaTimeStamp = timestamp - this.prevTimestamp;
   // timestamp is in milliseconds
   var dt = deltaTimeStamp / 1000.0;

   // update time center if isPlaying
   if(this.params.isPlaying)
   {
      const targetFps = 24;
      let targetFrameTime = 1.0 / 24;

      this.params.realTimeSinceLastPlayedFrame += dt;
      if(this.params.realTimeSinceLastPlayedFrame >= targetFrameTime)
      {
         var frameTime = this.params.realTimeSinceLastPlayedFrame;
         var normalizedTimeCenter = this.params.timeCenter / this.params.maxTimeCenter;
         var speed = 1.0 / this.params.reachMaxTimeSeconds;
         normalizedTimeCenter += speed * frameTime;
         normalizedTimeCenter -= Math.floor(normalizedTimeCenter);
         console.log(this.params.timeCenter);
         this.setTimeCenter(normalizedTimeCenter);

         this.params.realTimeSinceLastPlayedFrame = 0;
      }
   }

   var variablePerFrameVertexCount = true;
   // adjust perFrameVertexCount
   if (variablePerFrameVertexCount)
   {
      const targetFps = 24;
      let deltaTimes = this.params.deltaTimes;
      deltaTimes.push(dt);
      if(deltaTimes.length > 5)
      {
         var avgTime = 0;
         for(var i = 0; i < deltaTimes.length; ++i)
         {
            avgTime += deltaTimes[i]
         }
         avgTime /= deltaTimes.length;
         var avgFps = 1.0 / (avgTime + 0.00000001);
         if(avgFps > targetFps)
         {
            this.params.perFrameVertexCount = Math.min(this.params.maxPerFrameVertexCount, this.params.perFrameVertexCount + 100);
         }
         else if(avgFps < targetFps)
         {
            this.params.perFrameVertexCount = Math.max(this.params.minPerFrameVertexCount, this.params.perFrameVertexCount - 100);
         }
         deltaTimes = [this.deltaTimes[this.deltaTimes.length - 1]]

         // console.log(this.params.perFrameVertexCount)
      }
   }

   this.tracePhotons();
   this.updateVbos();
   this.prevTimestamp = timestamp;
   if (this.getTotalPathCount() == this.getTracedPathCount() && this.lastFinishedTimestamp == 0)
   {
      this.lastFinishedTimestamp = timestamp;
   }
}

PhotonRenderer.prototype.getContextBlitFactor = function(context)
{
   // var pathsTraced = this.params.singlePhotonMode.isEnabled ? 1 : this.params.tracedPathCount;
   var factor = 1.0;
   factor *= this.exposure;

   var renderer = context.renderer;
   if(this.params.isTransientEnabled)
   {
      factor *= renderer.getTransientExposure();
   }
   return factor;
};

PhotonRenderer.prototype.render = function(timestamp)
{
   var gl = this.gl;

   // first, render to context's frame buffer (sum buffer), without average
   var meanShader = this.meanShader;
   var params = this.params;
   var quadVbo = this.quadVbo;
   this.forEachContext(function(context){
      if (this.renderSingleContext && context != this.activeContext)
      {
         return;
      }

      var renderer = context.renderer;
      var fbo = context.fbo;
      fbo.bind();
      renderer.render(fbo);

      if (params.lastTracedPathCount == 0)
      {
         return;
      }

      var frontBufferIdx = context.frontBufferIdx;
      var backBufferIdx = 1 - frontBufferIdx;

      var oldAvg = context.avgBuffers[frontBufferIdx];
      var newSum = context.buffer;
      // render to backbuffer
      oldAvg.bind(0);
      newSum.bind(1);
      context.avgFbos[backBufferIdx].bind();

      meanShader.bind();
      meanShader.uniformTexture("u_OldAvg", oldAvg);
      meanShader.uniformTexture("u_NewSum", newSum);
      meanShader.uniform1F("u_TotalSampleCount", params.tracedPathCount);
      meanShader.uniform1F("u_NewSampleCount", params.lastTracedPathCount);

      quadVbo.bind();
      quadVbo.draw(meanShader, gl.TRIANGLE_STRIP);

      fbo.bind();
      fbo.clearColor(0, 0, 0, 0);

      context.frontBufferIdx = 1 - context.frontBufferIdx;
   }.bind(this))

   // switch to the active context
   var context = this.activeContext;
   var buffer = context.avgBuffers[context.frontBufferIdx];
   // render to window RT
   tgl.RenderTarget.unbind();

   // blit using the blitshader
   buffer.bind(0);
   this.blitShader.bind();
   this.blitShader.uniformTexture("u_Texture", buffer);

   var factor = this.getContextBlitFactor(context);
   this.blitShader.uniform1F("u_Factor", factor);

   this.quadVbo.bind();
   this.quadVbo.draw(this.blitShader, gl.TRIANGLE_STRIP);
}

PhotonRenderer.prototype.forEachRenderer = function(callback)
{
   for(var prim in this.primitiveContexts)
   {
      callback(this.primitiveContexts[prim].renderer);
   }
};

PhotonRenderer.prototype.forEachContext = function(callback)
{
   for(var prim in this.primitiveContexts)
   {
      callback(this.primitiveContexts[prim]);
   }
};

PhotonRenderer.prototype.updateVbos = function()
{
   // update the VBO in all children renderer
   this.forEachRenderer(function(renderer){
      if (this.renderSingleContext && this.activeContext && renderer != this.activeContext.renderer)
      {
         return;
      }
      renderer.updateVbo();
   }.bind(this));
}

PhotonRenderer.prototype.reset = function()
{
   this.lastResetTimestamp = this.prevTimestamp;
   this.lastFinishedTimestamp = 0;
   this.params.lastTracedPathCount = 0;
   this.params.lastTracedPhotonCount = 0;
   this.params.tracedPathCount = 0;
   this.params.tracedPhotonCount = 0;

   this.params.perFrameVertexCount = this.params.initPerFrameVertexCount;
   this.params.deltaTimes = [];

   this.forEachContext(function(context){
      var fbo = context.fbo;
      fbo.clearColor(0, 0, 0, 0);
   });

   this.params.tracer.reset();
   this.params.rng.reset();

   this.updateVbos();
};

PhotonRenderer.prototype.setCameraTheta = function(theta)
{
   this.forEachRenderer(function(renderer){
      renderer.setCameraTheta(theta);
   });
   this.reset();
};

// input t is from 0 to 1
PhotonRenderer.prototype.setTimeCenter = function(t)
{
   var center = t * this.params.maxTimeCenter;
   this.params.timeCenter = center;
   this.reset();
   return center;
};

PhotonRenderer.prototype.setTimeWidth = function(dt)
{
   this.params.timeWidth = dt;
   this.reset();
};

PhotonRenderer.prototype.setIsTransientEnabled = function(isTransientEnabled)
{
   this.params.isTransientEnabled = isTransientEnabled;
   this.reset();
};

PhotonRenderer.prototype.setIsPlaying = function(isPlaying)
{
   this.params.isPlaying = isPlaying;
   this.params.realTimeSinceLastPlayedFrame = 0.0;
};

PhotonRenderer.prototype.setPrimitive = function(newPrim, isPersp)
{
   if (isPersp)  newPrim += "_persp";

   this.activeContext = this.primitiveContexts[newPrim];

   var activeRenderer = null;
   for(let prim in this.primitiveContexts)
   {
      var renderer = this.primitiveContexts[prim].renderer;
      if (prim == newPrim)
      {
         activeRenderer = renderer;
      }
      else
      {
         renderer.setIsEnabled(false);
      }
   }
   activeRenderer.setIsEnabled(true);

   if (this.renderSingleContext)
   {
      this.reset();
   }
};

PhotonRenderer.prototype.setDrawSinglePhoton = function(enableDrawSingle, photonId)
{
   this.params.singlePhotonMode.isEnabled = enableDrawSingle;
   this.params.singlePhotonMode.photonId = photonId;
   this.reset();
};

PhotonRenderer.prototype.setExposure = function(exposure)
{
   this.exposure = exposure;
};

PhotonRenderer.prototype.setTotalPathCount = function(totalPathCount)
{
   this.params.totalPathCount = totalPathCount;
   this.reset();
};

PhotonRenderer.prototype.getProgress = function()
{
   return Math.min(this.params.tracedPathCount / this.params.totalPathCount, 1.0);
};

PhotonRenderer.prototype.getTotalPathCount = function()
{
   return this.params.totalPathCount;
};

PhotonRenderer.prototype.getTracedPathCount = function()
{
   return this.params.tracedPathCount;
};

PhotonRenderer.prototype.getSecondsElapsed = function()
{
   var endTimestamp = this.lastFinishedTimestamp == 0 ? this.prevTimestamp : this.lastFinishedTimestamp;
   var secondsElapsed = (endTimestamp - this.lastResetTimestamp) / 1000.0;
   return secondsElapsed;
};

PhotonRenderer.prototype.getPathPerSecond = function()
{
   var secondsElapsed = this.getSecondsElapsed();
   return this.params.tracedPathCount / secondsElapsed;
};

PhotonRenderer.prototype.saveAsPngImage = function()
{
   var link = document.createElement('a');
   link.setAttribute('download', 'transient-photon-primitive.png');
   link.href = this.gl.canvas.toDataURL('image/png');
   link.click();
   link.remove();
};

PhotonRenderer.prototype.setIsTimeWidthZeroSelectorEnabled = function(enabled)
{
   if (enabled)
   {
      this.isTimeWidthZeroSelector.parent.style.display = '';
   }
   else
   {
      this.isTimeWidthZeroSelector.parent.style.display = 'none';
   }
};

PhotonRenderer.prototype.saveAsImage = function()
{
   var factor = this.getContextBlitFactor(this.activeContext);
   var buffer = this.activeContext.avgBuffers[this.activeContext.frontBufferIdx];
   var imgFloatArray = buffer.toArray(factor);
   // NOTE: Warning: saving as monochrome image for now. (Magic word is 'Pf')
   // NOTE: Warning: assuming little endian (-1.0 instead of 1.0)
   var header = "Pf\n" + buffer.width + ' ' + buffer.height + '\n' + '-1.0' + '\n';

   // Flip the imgFloatArray vertically to conform with Photoshop.
   // (Pixels with each row are ordered left to right, rows ordered top to bottom.)
   for (var rowId = 0; rowId < buffer.height / 2; ++rowId)
   {
      var oppoRowId = buffer.height - 1 - rowId;
      for (var colId = 0; colId < buffer.width; ++colId)
      {
         var thisId = rowId * buffer.width + colId;
         var oppoId = oppoRowId * buffer.width + colId;
         var temp = imgFloatArray[thisId];
         imgFloatArray[thisId] = imgFloatArray[oppoId];
         imgFloatArray[oppoId] = temp;
      }
   }

   var blob = new Blob([header, imgFloatArray], {type: 'application/octet-stream'});

   // Ref: https://stackoverflow.com/questions/21012580/is-it-possible-to-write-data-to-file-using-only-javascript
   // If we are replacing a previously generated file we need to
   // manually revoke the object URL to avoid memory leaks.
   if (this.imageFile !== null) {
      window.URL.revokeObjectURL(this.imageFile);
   }

   this.imageFile = window.URL.createObjectURL(blob);

   var link = document.createElement('a');
   link.setAttribute('download', 'transient-photon-primitive.pfm');
   link.href = this.imageFile;
   link.click();

   link.remove();
};

var PrimitiveRenderer = function(photonRenderer, isPerspective = false)
{
   this.photonRenderer = photonRenderer;
   this.gl      = photonRenderer.gl;
   this.params  = photonRenderer.params;

   this.quadVbo = RenderUtilities.getQuadVbo();
   this.numRenderablePhotons = 0;

   // 2 triangles, each triangle has 3 vertices
   this.NumVerticesPerPlane = 2 * 3;
   // 6 quads, each quad has 2 triangles, each triangle has 3 vertices
   this.NumVerticesPerVolume = 6 * 2 * 3;

   // camera
   if (isPerspective)
   {
      this.camera = this.params.perspCamera;
   }
   else
   {
      this.camera = this.params.camera;
   }

   this.camTheta = 45;
   this.updateCamera();

   this.prepare(this.params.tracer);
   this.updateVbo();

   this.isEnabled = false;
   this.isEnabledCallback();
};

PrimitiveRenderer.prototype.isEnabledCallback = function()
{
   if(this.isEnabled)
   {
      this.onEnable();
   }
   else
   {
      this.onDisable();
   }
}

PrimitiveRenderer.prototype.setIsEnabled = function(isEnabled)
{
   if(this.isEnabled != isEnabled)
   {
      if (this.isPerspective())
      {
         // var maxTimeCenter = renderer.params.maxTimeCenter;
         // timeCenterSlider.setValue(2.2 / maxTimeCenter);
         if(this.params.timeCenter < 1.7)
         {
            this.photonRenderer.setTimeCenter(1.7 / this.params.maxTimeCenter);
            this.photonRenderer.timeCenterSlider.setValue(1.7 / this.params.maxTimeCenter);
         }
      }

      this.isEnabled = isEnabled;
      this.isEnabledCallback();
   }
}

// PrimitiveRenderer::onEnable(): called when this renderer is selected
PrimitiveRenderer.prototype.onEnable = function()
{
};

// PrimitiveRenderer::onDisable(): called when this renderer is deselected
PrimitiveRenderer.prototype.onDisable = function()
{
};

PrimitiveRenderer.prototype.isPerspective = function()
{
   return this.camera.isPerspective();
}
PrimitiveRenderer.prototype.isOrthographic = function()
{
   return this.camera.isOrthographic();
}

PrimitiveRenderer.prototype.updateVbo = function()
{
   throw new Error("PrimitiveRenderer: need to implement updateVbo() in children classes");
};

PrimitiveRenderer.prototype.addQuad = function(vertices, v0, v1, v2, v3)
{
   vertices.push(v0.x, v0.y, v0.z);
   vertices.push(v1.x, v1.y, v1.z);
   vertices.push(v3.x, v3.y, v3.z);
   vertices.push(v1.x, v1.y, v1.z);
   vertices.push(v2.x, v2.y, v2.z);
   vertices.push(v3.x, v3.y, v3.z);
};

PrimitiveRenderer.prototype.addVolume = function(vertices, x0, x1, x2, x3) {
   var v0 = x0;
   var v1 = x1;
   var v5 = x2;
   var v6 = x3;

   var dx = v1.subtract(v0);
   var dy = v5.subtract(v1);
   var dz = v6.subtract(v5);

   var v2 = v0.add(dx).add(dz);
   var v3 = v0.add(dz);
   var v4 = v0.add(dy);
   var v7 = v0.add(dy).add(dz);

   this.addQuad(vertices, v0, v1, v5, v4);
   this.addQuad(vertices, v1, v2, v6, v5);
   this.addQuad(vertices, v3, v7, v6, v2);
   this.addQuad(vertices, v0, v4, v7, v3);
   this.addQuad(vertices, v0, v3, v2, v1);
   this.addQuad(vertices, v4, v5, v6, v7);
};

// assumes a0, a1, a2, a3 are arrays
// a0[0] -> x of a0, a0[1] -> y of a0, ...
// "flatten" the data in the attributes array
// NOTE: WARNING: the sequence push the vertex must align with addQuad
PrimitiveRenderer.prototype.addQuadAttribute = function(attributes, a0, a1, a2, a3)
{
   console.assert(a0.length == a1.length);
   console.assert(a0.length == a2.length);
   console.assert(a0.length == a3.length);

   var pushVertex = function(attrs, vertex)
   {
      for(var i = 0; i < vertex.length; ++i)
      {
         attrs.push(vertex[i]);
      }
   };

   pushVertex(attributes, a0);
   pushVertex(attributes, a1);
   pushVertex(attributes, a3);
   pushVertex(attributes, a1);
   pushVertex(attributes, a2);
   pushVertex(attributes, a3);
}

PrimitiveRenderer.prototype.getPhotonIdInterval = function()
{
   var startId = 0;
   var endId = this.params.lastTracedPhotonCount;

   var singlePhotonMode = this.params.singlePhotonMode;
   if(singlePhotonMode.isEnabled)
   {
      startId = singlePhotonMode.photonId;
      endId = singlePhotonMode.photonId + 1;
   }

   return [startId, endId];
};

PrimitiveRenderer.prototype.forEachValidPhotonId = function(photonIdCallback, minBounce, maxBounce = Number.MAX_SAFE_INTEGER)
{

   const [startId, endId] = this.getPhotonIdInterval();

   for(var i = startId; i < endId; ++i)
   {
      if(this.params.tracer.bounce[i] < minBounce) continue;
      if(this.params.tracer.bounce[i] >= maxBounce) continue;
      photonIdCallback(i);
   }
};

PrimitiveRenderer.prototype.getTransientDistInterval = function()
{
   var isTimeWidthZero = this.supportsTimeWidthZero() && this.params.isTimeWidthZero;
   var width = isTimeWidthZero ? 0.0 : this.params.timeWidth;

   var timeBeg = this.params.timeCenter - width / 2.0;
   var timeEnd = this.params.timeCenter + width / 2.0;
   var distBeg = timeBeg * this.params.speedOfLight;
   var distEnd = timeEnd * this.params.speedOfLight;

   return [distBeg, distEnd];
};

PrimitiveRenderer.prototype.supportsTimeWidthZero = function()
{
   return false;
}

PrimitiveRenderer.prototype.getTransientExposure = function()
{
   if (this.supportsTimeWidthZero() && this.params.isTimeWidthZero)
   {
      return 1.0;
   }
   else
   {
      return 1.0 / (this.params.timeWidth * this.params.speedOfLight);
   }
};


PrimitiveRenderer.prototype.setCameraTheta = function(theta)
{
   this.camTheta = theta;
   this.updateCamera();
};

PrimitiveRenderer.prototype.updateCamera = function()
{
   // projection matrix
   var proj = this.camera.getProjectionMatrix();

   // view matrix
   var camToCenterDist = 1.5;
   var deg2Rad = function(deg)
   {
      return deg * (Math.PI / 180);
   };

   var eyePos = new Vector(camToCenterDist * Math.sin(deg2Rad(this.camTheta)), 0.0, camToCenterDist * Math.cos(deg2Rad(this.camTheta)));
   // eyePos = new Vector(0, camToCenterDist, 0);
   // eyePos = new Vector(camToCenterDist * Math.sin(deg2Rad(this.camTheta)), 0.75, camToCenterDist * Math.cos(deg2Rad(this.camTheta)));

   var centerPos = new Vector(0, 0, 0);
   var upDir = new Vector(0, 1, 0);

   this.camera.setTransform(eyePos, centerPos, upDir);
   var viewMat = this.camera.getTransform();

   // output variables to prototype
   // this.eyePos = eyePos;
   this.viewMat = viewMat;
   this.projMat = proj;
   // mvp matrix
   this.mvp = Matrix.multiply(proj, viewMat);
};

PrimitiveRenderer.prototype.evalCamPhase = function(wi)
{
   if (!this.params.evalCamPhase)
   {
      return 1.0;
   }
   else if(this.camera.isOrthographic())
   {
      var camToWorldMat = this.viewMat;
      var camDir = new Vector(camToWorldMat.m[2], camToWorldMat.m[6], camToWorldMat.m[10]);
      return this.params.tracer.evalPhase(wi, camDir);
   }
   else
   {
      throw new Error("evalCamPhase() does not support perspective camera.");
   }
}
var Primitive2DRenderer = function(photonRenderer, isPerspective = false)
{
   PrimitiveRenderer.call(this, photonRenderer, isPerspective);
};

// inheritance: Primitive2DRenderer -> PrimitiveRenderer
Primitive2DRenderer.prototype = Object.create(PrimitiveRenderer.prototype);
Primitive2DRenderer.prototype.constructor = Primitive2DRenderer;

Primitive2DRenderer.prototype.prepare = function(tracer)
{
   var gl = this.gl;

   // beam can reuse the shader for plane
   this.shader = new tgl.Shader("plane-vert.shader", "plane-frag.shader");
   this.transientShader = new tgl.Shader("plane-vert.shader", "plane-frag.shader", {'ENABLE_TRANSIENT': 1});

   this.vbo = new tgl.VertexBuffer();
   this.vbo.addAttribute('a_Position', gl.FLOAT, 3);
   this.vbo.addAttribute('a_Transform', gl.FLOAT, 3, 3, 1);
   this.vbo.addAttribute('a_Offset', gl.FLOAT, 3, 1, 1);
   this.vbo.addAttribute('a_Color', gl.FLOAT, 3, 1, 1);
   this.vbo.addAttribute('a_InitDist', gl.FLOAT, 1, 1, 1);
   this.vbo.addAttribute('a_PlanePos', gl.FLOAT, 2, 1, 1);
   if (this.isPerspective())
   {
      this.vbo.addAttribute('a_PhotonDirection', gl.FLOAT, 3, 1, 1);
      this.vbo.addAttribute('a_PlaneNormal', gl.FLOAT, 3, 1, 1);
   }
};

Primitive2DRenderer.prototype.addPhotonPrimitive2D = function(vboData, tracer, lastPhotonId, camToWorldMat)
{
   throw new Error("Primitive2DRenderer: need to implement updateVbo() in children classes");
};

Primitive2DRenderer.prototype.minLightBounces = function()
{
   throw new Error("Primitive2DRenderer: need to implement minLightBounces() in children classes");
};

Primitive2DRenderer.prototype.updateVbo = function()
{
   var vboData = {};
   vboData['position'] = [
       0,  0,  0,
       1,  0,  0,
       0,  1,  0,
       1,  1,  0,
   ];
   vboData['transform'] = [];
   vboData['offset'] = [];
   vboData['color'] = [];
   vboData['initDist'] = [];
   vboData['planePos'] = [];
   if (this.isPerspective())
   {
      vboData['photonDirection'] = [];
      vboData['planeNormal'] = [];
   }

   var numRenderablePhotons = 0;
   this.forEachValidPhotonId(function(i){
      numRenderablePhotons++;

      this.addPhotonPrimitive2D(vboData, this.params.tracer, i, this.viewMat.inverse());
   }.bind(this), Math.max(this.minLightBounces(), this.params.numLightBounces), this.params.numLightBounces + 1);

   this.vbo.init(4, numRenderablePhotons);
   this.vbo.copyAttribute('a_Position' , new Float32Array(vboData['position']));
   this.vbo.copyAttribute('a_Transform', new Float32Array(vboData['transform']));
   this.vbo.copyAttribute('a_Offset', new Float32Array(vboData['offset']));
   this.vbo.copyAttribute('a_Color', new Float32Array(vboData['color']));
   this.vbo.copyAttribute('a_InitDist', new Float32Array(vboData['initDist']));
   this.vbo.copyAttribute('a_PlanePos', new Float32Array(vboData['planePos']));
   if (this.isPerspective())
   {
      this.vbo.copyAttribute('a_PhotonDirection', new Float32Array(vboData['photonDirection']));
      this.vbo.copyAttribute('a_PlaneNormal', new Float32Array(vboData['planeNormal']));
   }

   this.numRenderablePhotons = numRenderablePhotons;
};

Primitive2DRenderer.prototype.render = function()
{
   var gl = this.gl;

   var activeShader = this.params.isTransientEnabled ? this.transientShader : this.shader;
   this.vbo.bind();

   gl.disable(gl.CULL_FACE);
   gl.enable(gl.BLEND);
   gl.blendFunc(gl.ONE, gl.ONE);

   activeShader.bind();
   activeShader.uniformMatrix('u_ModelViewProjection', this.mvp);

   if(this.params.isTransientEnabled)
   {
      // calculate the longest and shortest photon path distance
      // for the time interval we are considering
      const [distBeg, distEnd] = this.getTransientDistInterval();
      activeShader.uniform2F("u_DistInterval", distBeg, distEnd);
   }

   activeShader.uniformMatrix('u_ModelView', this.viewMat);
   activeShader.uniform1B('u_EvalCamTransmittance', this.params.evalCamTransmittance);
   activeShader.uniform3F('u_SigmaT',this.params.sigmaT.x, this.params.sigmaT.y, this.params.sigmaT.z);
   activeShader.uniform1B('u_IsPerspective', this.isPerspective());

   if (this.isPerspective())
   {
      activeShader.uniform1B('u_CameraWarped', true);
   }

   this.vbo.bind();
   this.vbo.draw(activeShader, gl.TRIANGLE_STRIP);

   gl.disable(gl.BLEND);
};

var BeamRenderer = function(photonRenderer, isPerspective = false)
{
   Primitive2DRenderer.call(this, photonRenderer, isPerspective);
};

// inheritance: BeamRenderer -> Primitive2DRenderer
BeamRenderer.prototype = Object.create(Primitive2DRenderer.prototype);
BeamRenderer.prototype.constructor = BeamRenderer;

BeamRenderer.prototype.prepare = function(tracer)
{
   Primitive2DRenderer.prototype.prepare.call(this, tracer);

   // setup UI
   var beamRenderer = this;
   var sliderName = this.isOrthographic() ? "blur-radius" : "blur-radius-persp";
   this.blurRadiusSlider = new tui.Slider(sliderName, 0.01, 0.1, true, false, function(r){
      beamRenderer.blurRadius = r;
      beamRenderer.photonRenderer.reset();
      this.setLabel("blur radius: " + Utilities.formatFloat(r));
   });
   this.blurRadiusSlider.setValue(0.01);
};

BeamRenderer.prototype.onEnable = function()
{
   this.blurRadiusSlider.parent.style.display = '';
};

BeamRenderer.prototype.onDisable = function()
{
   this.blurRadiusSlider.parent.style.display = 'none';
};

// override: addPhotonPrimitive2D()
BeamRenderer.prototype.addPhotonPrimitive2D = function(vboData, tracer, lastPhotonId, camToWorldMat)
{
   var i = lastPhotonId;
   var v0 = new Vector(tracer.posX[i - 1], tracer.posY[i - 1], tracer.posZ[i - 1]);
   var v1 = new Vector(tracer.posX[i - 0], tracer.posY[i - 0], tracer.posZ[i - 0]);
   var color = new Vector(tracer.colorR[i - 0], tracer.colorG[i - 0], tracer.colorB[i - 0]);
   var initDist = tracer.pathDistance[i - 1];

   this.addPhotonBeam(vboData, v0, v1, color, camToWorldMat, initDist);
};

BeamRenderer.prototype.addPhotonBeam = function(vboData, v0, v1, color, camToWorldMat, initDist)
{
   var beamDir = v1.subtract(v0).unit();
   var camDir = new Vector(-camToWorldMat.m[2], -camToWorldMat.m[6], -camToWorldMat.m[10]);
   var perpDir = beamDir.cross(camDir.negative());
   var sinTheta = Math.sqrt(perpDir.dot(perpDir));

   var hPerp = perpDir.multiply(this.blurRadius / sinTheta);

   var offset = v0.subtract(hPerp);
   var xTransform = v1.subtract(v0);
   var yTransform = hPerp.multiply(2.0);
   var zTransform = xTransform.cross(yTransform).unit();

   vboData['transform'].push(xTransform.x, xTransform.y, xTransform.z);
   vboData['transform'].push(yTransform.x, yTransform.y, yTransform.z);
   vboData['transform'].push(zTransform.x, zTransform.y, zTransform.z);

   vboData['offset'].push(offset.x, offset.y, offset.z);

   //// initDist: 1D, the initial distance of the photon plane
   // i.e. how for the path has traveled before hitting reaching the photon at id i-2
   // it is the same across all photon plane vertices
   vboData['initDist'].push(initDist);

   if (this.isPerspective())
   {
      var planeNormal = perpDir.unit().cross(beamDir);
      vboData['photonDirection'].push(beamDir.x, beamDir.y, beamDir.z);
      vboData['planeNormal'].push(planeNormal.x, planeNormal.y, planeNormal.z);
   }


   //// planePos: 2D, the position of the vertex in "plane-local" space
   //          The x dir in this space is (v1 - v0) in world space. One unit in this dir is one unit in world space.
   //          planeDist.x: how far the path traveled in dir (v1 - v0), in world space
   // NOTE: one of the axis is wasted when rendering beams.
   // TODO: optimize this to only use one float for each quad
   var dx = v1.subtract(v0);
   var distX = dx.length();
   vboData['planePos'].push(distX, 0);

   color = color.clone();
   // blur normalization
   color = color.multiply(1.0 / (2 * this.blurRadius));
   // cancel last free path sampling
   color = color.multiply(this.params.sigmaT);

   if (this.isOrthographic())
   {
      // calculate jacobian
      color = color.divide(sinTheta);
      // the phase function for the scatter connecting cam path to photon path
      // NOTE: WARNING: only works for orthographic cameras
      color = color.multiply(this.evalCamPhase(beamDir.negative()));
   }

   vboData['color'].push(color.x, color.y, color.z);
};

BeamRenderer.prototype.minLightBounces = function()
{
   return 1;
}

var PlaneRenderer = function(photonRenderer, isPerspective)
{
   Primitive2DRenderer.call(this, photonRenderer, isPerspective);
};

// inheritance: PlaneRenderer -> Primitive2DRenderer
PlaneRenderer.prototype = Object.create(Primitive2DRenderer.prototype);
PlaneRenderer.prototype.constructor = PlaneRenderer;

// override: Primitive2DRenderer::addPhotonPrimitive2D
PlaneRenderer.prototype.addPhotonPrimitive2D = function(vboData, tracer, lastPhotonId, camToWorldMat)
{
   var i = lastPhotonId;

   var v0 = new Vector(tracer.posX[i - 2], tracer.posY[i - 2], tracer.posZ[i - 2]);
   var v1 = new Vector(tracer.posX[i - 1], tracer.posY[i - 1], tracer.posZ[i - 1]);
   var v2 = new Vector(tracer.posX[i - 0], tracer.posY[i - 0], tracer.posZ[i - 0]);
   var color = new Vector(tracer.colorR[i - 0], tracer.colorG[i - 0], tracer.colorB[i - 0]);
   var initDist = tracer.pathDistance[i - 2];

   this.addPhotonPlane(vboData, v0, v1, v2, color, camToWorldMat, initDist);
};

PlaneRenderer.prototype.addPhotonPlane = function(vboData, v0, v1, v2, color, camToWorldMat, initDist)
{
   var dx = v1.subtract(v0);
   var dy = v2.subtract(v1);
   var dz = dx.cross(dy).unit();

   vboData['transform'].push(dx.x, dx.y, dx.z);
   vboData['transform'].push(dy.x, dy.y, dy.z);
   vboData['transform'].push(dz.x, dz.y, dz.z);

   vboData['offset'].push(v0.x, v0.y, v0.z);

   // initDist: 1D, the initial distance of the photon plane
   // i.e. how for the path has traveled before hitting reaching the photon at id i-2
   // it is the same across all photon plane vertices
   vboData['initDist'].push(initDist);

   // planePos: 2D, the position of the vertex in "plane-local" space
   //          The x dir in this space is (v1 - v0) in world space. One unit in this dir is one unit in world space.
   //          The y dir in this space is (v2 - v1) in world space. One unit in this dir is one unit in world space.
   //          planeDist.x: how far the path traveled in dir (v1 - v0), in world space
   //          planeDist.y: how far the path traveled in dir (v2 - v1), in world space

   var distX = dx.length();
   var distY = dy.length();
   vboData['planePos'].push(distX, distY);

   if (this.isPerspective())
   {
      var planeNormal = dx.unit().cross(dy.unit());
      vboData['photonDirection'].push(dy.unit().x, dy.unit().y, dy.unit().z);
      vboData['planeNormal'].push(planeNormal.x, planeNormal.y, planeNormal.z);
   }

   var camDir = new Vector(camToWorldMat.m[2], camToWorldMat.m[6], camToWorldMat.m[10]);
   var determinant = Math.abs(dx.dot(dy.cross(camDir))) / (distX * distY);
   var determinantP = Math.abs(dx.unit().dot(dy.unit().cross(camDir)));
   if(!(Math.abs(determinant - determinantP) < 0.001))
   {
      throw new Error("determinant");
   }
   console.assert(Math.abs(determinant - determinantP) < 0.001);

   // For each f(t_i) showed up in the photon volume estimator,
   // we need to time that by sigmaT to "cancel out" the original C(t_i) = f(t_i) / p(t_i)
   // factor already calculated in the photon's color.
   // Note: f(t_i) / p(t_i) = \frac{1}{\sigma_t}, since we used free path sampling

   // In photon planes, f(t_i) shows up two times f(t_{l-1}), f(t_{l})),
   // thus we need to multiply three times
   color = color.clone();
   for(var i = 0; i < 2; ++i)
   {
      color = color.multiply(this.params.tracer.sigmaT);
   }

   if (this.isOrthographic())
   {
      // determinant
      color = color.divide(determinant);
      // the phase function for the scatter connecting cam path to photon path
      // NOTE: WARNING: only works for orthographic cameras
      color = color.multiply(this.evalCamPhase(dy.unit().negative()));
   }


   vboData['color'].push(color.x, color.y, color.z);
};

PlaneRenderer.prototype.minLightBounces = function()
{
   return 2;
}


var VolumeRenderer = function(photonRenderer, isPerspective = false)
{
   PrimitiveRenderer.call(this, photonRenderer, isPerspective);
};

// inheritance: VolumeRenderer -> Primitive Renderer
VolumeRenderer.prototype = Object.create(PrimitiveRenderer.prototype);
VolumeRenderer.prototype.constructor = VolumeRenderer;

VolumeRenderer.prototype.isVolumeValid = function(v0, v1, v5, v6) {
   var dx = v1.subtract(v0);
   var dy = v5.subtract(v1);
   var dz = v6.subtract(v5);

   var determinant = Math.abs(dx.dot(dy.cross(dz))) / (dx.length() * dy.length() * dz.length());

   if (determinant < 1E-4) {
      return false;
   } else {
      return true;
   }
};



VolumeRenderer.prototype.addPhotonVolume = function (x0, x1, x2, x3) {
   this.addVolume(this.positionVboData, x0, x1, x2, x3)

   var dx = x1.subtract(x0);
   var dy = x2.subtract(x1);
   var dz = x3.subtract(x2);

   var addVolumeMat = function(mats, e0, e1, e2, offset)
   {
      // opengl stores matrix in a column-major way
      mats.push(e0.x, e0.y, e0.z, 0);
      mats.push(e1.x, e1.y, e1.z, 0);
      mats.push(e2.x, e2.y, e2.z, 0);
      mats.push(offset.x, offset.y, offset.z, 1);
   };

   // for now, add a volume matrix to each vertex
   // TODO: reduce # of params passed to the GPU with HW instancing
   for(var j = 0; j < this.NumVerticesPerVolume; ++j)
   {
      addVolumeMat(this.volumeMatVboData, dx, dy, dz, x0);
   }
};

VolumeRenderer.prototype.addPhotonColor = function(v0, v1, v5, v6, r, g, b) {
   var dx = v1.subtract(v0);
   var dy = v5.subtract(v1);
   var dz = v6.subtract(v5);


   var determinant = Math.abs(dx.dot(dy.cross(dz))) / (dx.length() * dy.length() * dz.length());
   var color = new Vector(r, g, b);

   // For each f(t_i) showed up in the photon volume estimator,
   // we need to time that by sigmaT to "cancel out" the original C(t_i) = f(t_i) / p(t_i)
   // factor already calculated in the photon's color.
   // Note: f(t_i) / p(t_i) = \frac{1}{\sigma_t}, since we used free path sampling

   // In photon volumes, f(t_i) shows up three times (f(t_{l-2}), f(t_{l-1}), f(t_{l})),
   // thus we need to multiply three times
   for(var i = 0; i < 3; ++i)
   {
      color = color.multiply(this.params.tracer.sigmaT);
   }

   // determinant
   color = color.divide(determinant);

   for(var j = 0; j < this.NumVerticesPerVolume; ++j)
   {
      this.colorVboData.push(color.x, color.y, color.z);
   }
}

VolumeRenderer.prototype.prepare = function(tracer)
{
   var gl = this.gl;

   this.volumeShader = new tgl.Shader("volume-vert.shader", "volume-frag.shader");
   this.transientShader = new tgl.Shader("volume-vert.shader", "volume-frag.shader", {'ENABLE_TRANSIENT': 1});

   this.vbo = new tgl.VertexBuffer();
   this.vbo.addAttribute('a_Position', gl.FLOAT, 3);
   this.vbo.addAttribute('a_VolumeMatrix', gl.FLOAT, 4, 4);
   this.vbo.addAttribute('a_Color', gl.FLOAT, 3);
   this.vbo.addAttribute('a_DistInterval', gl.FLOAT, 2);
};

VolumeRenderer.prototype.onEnable = function()
{
   if (!this.isPerspective())
      this.photonRenderer.setIsTimeWidthZeroSelectorEnabled(true);
};

VolumeRenderer.prototype.onDisable = function()
{
   if (!this.isPerspective())
      this.photonRenderer.setIsTimeWidthZeroSelectorEnabled(false);
};

VolumeRenderer.prototype.updateVbo = function()
{
   var gl = this.gl;
   var vbo = this.vbo;

   // cpu struct of arrays for volume vertices
   this.positionVboData = [];
   // a SoA describing the matrix that transform a point in unit cube
   // to a point in world space inside the photon volume.
   // e0: corner 0 of cube, using the photon volume's origin as origin
   // e1: corner 1 of cube, using the photon volume's origin as origin
   // e2: corner 2 of cube, using the photon volume's origin as origin
   // offset: offset from photon volume's origin to the world space origin

   // e0 = v0 - o
   // e1 = v1 - o
   // e2 = v2 - o
   // offset = o - ow

   //      v1 _____
   //        /    /|
   //     o /____/v2
   //       |    | |
   // ow    |____|/
   //     v0
   this.volumeMatVboData = [];
   // color for each photon volume
   this.colorVboData = [];

   this.numRenderablePhotons = 0;
   // for each photon, build 6 square faces (12 triangle faces) to render the
   // photon volume
   this.forEachValidPhotonId(function(i){
      var v0 = new Vector(this.params.tracer.posX[i - 3], this.params.tracer.posY[i - 3], this.params.tracer.posZ[i - 3]);
      var v1 = new Vector(this.params.tracer.posX[i - 2], this.params.tracer.posY[i - 2], this.params.tracer.posZ[i - 2]);
      var v2 = new Vector(this.params.tracer.posX[i - 1], this.params.tracer.posY[i - 1], this.params.tracer.posZ[i - 1]);
      var v3 = new Vector(this.params.tracer.posX[i - 0], this.params.tracer.posY[i - 0], this.params.tracer.posZ[i - 0]);

      if(this.isVolumeValid(v0, v1, v2, v3))
      {
         this.numRenderablePhotons++;
         this.addPhotonVolume(v0, v1, v2, v3);
         this.addPhotonColor(
            v0, v1, v2, v3,
            this.params.tracer.colorR[i - 0], this.params.tracer.colorG[i - 0], this.params.tracer.colorB[i - 0]
         );
      }
   }.bind(this), Math.max(3, this.params.numLightBounces), this.params.numLightBounces + 1);

   vbo.init(this.numRenderablePhotons * this.NumVerticesPerVolume);
   console.assert(this.positionVboData.length == vbo.numVertices * 3);

   vbo.copyAttribute('a_Position', new Float32Array(this.positionVboData));
   vbo.copyAttribute('a_VolumeMatrix', new Float32Array(this.volumeMatVboData));
   vbo.copyAttribute('a_Color', new Float32Array(this.colorVboData));

   this.updateTransientVbo();
};


// return value: a tgl.Texture instance that contains the result of render (without final blit)
VolumeRenderer.prototype.render = function(outFbo)
{
   var gl = this.gl;
   var activeShader = this.params.isTransientEnabled ? this.transientShader : this.volumeShader;

   activeShader.bind();
   activeShader.uniformMatrix('u_ModelView', this.viewMat);
   activeShader.uniformMatrix('u_ModelViewProjection', this.mvp);
   activeShader.uniformMatrix('u_View', this.viewMat);
   activeShader.uniform2F('u_Resolution', this.params.width, this.params.height);
   activeShader.uniformMatrix('u_InvProjection', this.projMat.inverse());

   activeShader.uniform1B('u_EvalCamTransmittance', this.params.evalCamTransmittance);
   activeShader.uniform3F('u_SigmaT',this.params.sigmaT.x, this.params.sigmaT.y, this.params.sigmaT.z);

   if (this.isPerspective())
   {
      activeShader.uniform1B('u_CameraWarped', true);
   }

   this.vbo.bind();

   gl.enable(gl.BLEND);
   gl.blendFunc(gl.ONE, gl.ONE);
   gl.enable(gl.CULL_FACE);
   gl.cullFace(gl.BACK);

   this.vbo.draw(activeShader, gl.TRIANGLES);

   gl.disable(gl.CULL_FACE);
   gl.disable(gl.BLEND);
};

VolumeRenderer.prototype.updateTransientVbo = function()
{
   var distInterval = this.getTransientDistInterval();
   var distBeg = distInterval[0];
   var distEnd = distInterval[1];

   var distIntervalVboData = [];

   this.forEachValidPhotonId(function(i){
      var v0 = new Vector(this.params.tracer.posX[i - 3], this.params.tracer.posY[i - 3], this.params.tracer.posZ[i - 3]);
      var v1 = new Vector(this.params.tracer.posX[i - 2], this.params.tracer.posY[i - 2], this.params.tracer.posZ[i - 2]);
      var v2 = new Vector(this.params.tracer.posX[i - 1], this.params.tracer.posY[i - 1], this.params.tracer.posZ[i - 1]);
      var v3 = new Vector(this.params.tracer.posX[i - 0], this.params.tracer.posY[i - 0], this.params.tracer.posZ[i - 0]);

      if(!this.isVolumeValid(v0, v1, v2, v3)) return;

      var startDist = this.params.tracer.pathDistance[i - 3];

      var volumeDistBeg = distBeg - startDist;
      var volumeDistEnd = distEnd - startDist;

      for(var j = 0; j < this.NumVerticesPerVolume; ++j)
      {
         distIntervalVboData.push(volumeDistBeg, volumeDistEnd);
      }
   }.bind(this), Math.max(3, this.params.numLightBounces), this.params.numLightBounces + 1);

   this.vbo.copyAttribute('a_DistInterval', new Float32Array(distIntervalVboData));
};

VolumeRenderer.prototype.supportsTimeWidthZero = function()
{
   return !this.isPerspective();
}

var EllipsoidRenderer = function(photonRenderer)
{
   PrimitiveRenderer.call(this, photonRenderer, true);
};

// inheritance: EllipsoidRenderer -> PrimitiveRenderer
EllipsoidRenderer.prototype = Object.create(PrimitiveRenderer.prototype);
EllipsoidRenderer.prototype.constructor = EllipsoidRenderer;

EllipsoidRenderer.prototype.prepare = function(tracer)
{
   console.assert(this.isPerspective());
   var gl = this.gl;

   this.shader = new tgl.Shader("ellipsoid-vert.shader", "ellipsoid-frag.shader", undefined, true);

   this.vbo = new tgl.VertexBuffer();
   this.vbo.addAttribute('a_Position', gl.FLOAT, 2);
   this.vbo.addAttribute('a_FocalPos', gl.FLOAT, 3, 1, 1);
   this.vbo.addAttribute('a_MinTau', gl.FLOAT, 1, 1, 1);
   this.vbo.addAttribute('a_MaxTau', gl.FLOAT, 1, 1, 1);
   this.vbo.addAttribute('a_Color', gl.FLOAT, 3, 1, 1);
};

EllipsoidRenderer.prototype.addEllipsoid = function(vboData, tracer, i, camPos, viewMat)
{
   var v0 = new Vector(tracer.posX[i - 0], tracer.posY[i - 0], tracer.posZ[i - 0]);
   var v1 = new Vector(tracer.posX[i - 1], tracer.posY[i - 1], tracer.posZ[i - 1]);
   var color = new Vector(tracer.colorR[i - 0], tracer.colorG[i - 0], tracer.colorB[i - 0]);
   var lightPathDist = tracer.pathDistance[i - 0];

   var f = camPos.subtract(v0).length();
   const [distBeg, distEnd] = this.getTransientDistInterval();
   var tau1 = distBeg - lightPathDist;
   var tau2 = distEnd - lightPathDist;

   if (tau1 < f && tau2 < f)
   {
      return false;
   }
   var tau1 = Math.max(tau1, f);
   var tau2 = Math.max(tau2, f);
   console.assert(tau1 >= f);
   console.assert(tau2 >= f);
   console.assert(tau2 >= tau1);
   if (this.params.isTimeWidthZero)
   {
      console.assert(tau1 == tau2);
   }

   var v0View = viewMat.transformPoint(v0);
   vboData['focalPos'].push(v0View.x, v0View.y, v0View.z);
   vboData['minTau'].push(tau1);
   vboData['maxTau'].push(tau2);
   vboData['color'].push(color.x, color.y, color.z);

   return true;
};

EllipsoidRenderer.prototype.updateVbo = function()
{
   var vboData = {};
   vboData['position'] = [
      -1, -1,
       1, -1,
      -1,  1,
       1,  1,
   ];
   vboData['focalPos'] = [];
   vboData['minTau'] = [];
   vboData['maxTau'] = [];
   vboData['color'] = [];

   var camPos = this.camera.getPosition();
   var numRenderablePhotons = 0;
   this.forEachValidPhotonId(function(i){
      if (this.addEllipsoid(vboData, this.params.tracer, i, camPos, this.viewMat))
      {
         numRenderablePhotons++;
      }
   }.bind(this), Math.max(0, this.params.numLightBounces - 1), this.params.numLightBounces);

   this.vbo.init(4, numRenderablePhotons);

   this.vbo.copyAttribute('a_Position', new Float32Array(vboData['position']));
   this.vbo.copyAttribute('a_FocalPos' , new Float32Array(vboData['focalPos']));
   this.vbo.copyAttribute('a_MinTau', new Float32Array(vboData['minTau']));
   this.vbo.copyAttribute('a_MaxTau', new Float32Array(vboData['maxTau']));
   this.vbo.copyAttribute('a_Color', new Float32Array(vboData['color']));

   this.numRenderablePhotons = numRenderablePhotons;
};

EllipsoidRenderer.prototype.render = function()
{
   var gl = this.gl;
   var activeShader = this.shader;

   activeShader.bind();

   var frustumCorners = this.camera.calculateViewSpaceFrustumCorners(1.0);
   activeShader.uniform3F("u_FrustumCornerBL", frustumCorners[0].x, frustumCorners[0].y, frustumCorners[0].z);
   activeShader.uniform3F("u_FrustumCornerBR", frustumCorners[1].x, frustumCorners[1].y, frustumCorners[1].z);
   activeShader.uniform3F("u_FrustumCornerTL", frustumCorners[3].x, frustumCorners[3].y, frustumCorners[3].z);
   activeShader.uniform3F("u_FrustumCornerTR", frustumCorners[2].x, frustumCorners[2].y, frustumCorners[2].z);
   activeShader.uniform1F("u_Time", this.params.timestamp / 1000.0);
   activeShader.uniform3F('u_SigmaT',this.params.sigmaT.x, this.params.sigmaT.y, this.params.sigmaT.z);
   activeShader.uniform3F('u_SigmaS',this.params.sigmaS.x, this.params.sigmaS.y, this.params.sigmaS.z);
   activeShader.uniform1B('u_IsTimeWidthZero', this.params.isTimeWidthZero);

   this.vbo.bind();

   gl.enable(gl.BLEND);
   gl.blendFunc(gl.ONE, gl.ONE);
   gl.disable(gl.CULL_FACE);

   this.vbo.draw(activeShader, gl.TRIANGLE_STRIP);

   gl.disable(gl.CULL_FACE);
   gl.disable(gl.BLEND);
};

EllipsoidRenderer.prototype.supportsTimeWidthZero = function()
{
   return true;
}

EllipsoidRenderer.prototype.onEnable = function()
{
   this.photonRenderer.setIsTimeWidthZeroSelectorEnabled(true);
};

EllipsoidRenderer.prototype.onDisable = function()
{
   this.photonRenderer.setIsTimeWidthZeroSelectorEnabled(false);
};


var SphereRenderer = function(photonRenderer)
{
   PrimitiveRenderer.call(this, photonRenderer, false);
};

// inheritance: Primitive2DRenderer -> PrimitiveRenderer
SphereRenderer.prototype = Object.create(PrimitiveRenderer.prototype);
SphereRenderer.prototype.constructor = SphereRenderer;

SphereRenderer.prototype.prepare = function(tracer)
{
   console.assert(!this.isPerspective());
   var gl = this.gl;

   this.shader = new tgl.Shader("sphere-vert.shader", "sphere-frag.shader", undefined, true);

   this.vbo = new tgl.VertexBuffer();
   this.vbo.addAttribute('a_Position', gl.FLOAT, 3);
   this.vbo.addAttribute('a_Offset', gl.FLOAT, 3, 1, 1);
   this.vbo.addAttribute('a_Radius', gl.FLOAT, 1, 1, 1);
   this.vbo.addAttribute('a_InDirection', gl.FLOAT, 3, 1, 1);
   this.vbo.addAttribute('a_Color', gl.FLOAT, 3, 1, 1);

};

SphereRenderer.prototype.addPhotonSphere = function(vboData, tracer, i, rng)
{
   var v0 = new Vector(tracer.posX[i - 0], tracer.posY[i - 0], tracer.posZ[i - 0]);
   var v1 = new Vector(tracer.posX[i - 1], tracer.posY[i - 1], tracer.posZ[i - 1]);
   var v2 = new Vector(tracer.posX[i - 2], tracer.posY[i - 2], tracer.posZ[i - 2]);

   var s1 = v0.subtract(v1).length();
   const [distBeg, distEnd] = this.getTransientDistInterval();
   var lightPathDist = tracer.pathDistance[i - 1];
   var sampleTotalDist = lerp(distBeg, distEnd, rng.next());
   if (lightPathDist > sampleTotalDist)
   {
      return false;
   }

   var radius = sampleTotalDist - lightPathDist;
   if (radius > s1)
   {
      return false;
   }

   var color = new Vector(tracer.colorR[i - 0], tracer.colorG[i - 0], tracer.colorB[i - 0]);
   if (distBeg != distEnd)
   {
      color = color.multiply(distEnd - distBeg);
   }

   var inDir = v1.subtract(v2).unit();

   vboData['offset'].push(v1.x, v1.y, v1.z);
   vboData['radius'].push(radius);
   vboData['inDirection'].push(inDir.x, inDir.y, inDir.z);
   vboData['color'].push(color.x, color.y, color.z);

   return true;
};

SphereRenderer.prototype.minLightBounces = function()
{
   return 2;
};

SphereRenderer.prototype.updateVbo = function()
{
   var vboData = {};
   vboData['position'] = [];
   this.addVolume(
      vboData['position'],
      new Vector(1, 1, -1),
      new Vector(1, 1, 1),
      new Vector(1, -1, 1),
      new Vector(-1, -1, 1)
   );
   vboData['offset'] = [];
   vboData['radius'] = [];
   vboData['inDirection'] = [];
   vboData['color'] = [];


   var numRenderablePhotons = 0;
   this.forEachValidPhotonId(function(i){
      if (this.addPhotonSphere(vboData, this.params.tracer, i, this.params.rng))
      {
         numRenderablePhotons++;
      }
   }.bind(this), Math.max(this.minLightBounces(), this.params.numLightBounces), this.params.numLightBounces + 1);

   this.vbo.init(this.NumVerticesPerVolume, numRenderablePhotons);
   this.vbo.copyAttribute('a_Position' , new Float32Array(vboData['position']));
   this.vbo.copyAttribute('a_Offset', new Float32Array(vboData['offset']));
   this.vbo.copyAttribute('a_Radius', new Float32Array(vboData['radius']));
   this.vbo.copyAttribute('a_InDirection', new Float32Array(vboData['inDirection']));
   this.vbo.copyAttribute('a_Color', new Float32Array(vboData['color']));

   this.numRenderablePhotons = numRenderablePhotons;
};

SphereRenderer.prototype.render = function()
{
   var gl = this.gl;
   var activeShader = this.shader;

   activeShader.bind();

   activeShader.uniformMatrix('u_ModelView', this.viewMat);
   activeShader.uniformMatrix('u_ModelViewProjection', this.mvp);
   activeShader.uniform2F('u_Resolution', this.params.width, this.params.height);
   activeShader.uniformMatrix('u_InvProjection', this.projMat.inverse());
   activeShader.uniform3F('u_SigmaT',this.params.sigmaT.x, this.params.sigmaT.y, this.params.sigmaT.z);

   this.vbo.bind();

   gl.enable(gl.BLEND);
   gl.blendFunc(gl.ONE, gl.ONE);
   gl.enable(gl.CULL_FACE);
   gl.cullFace(gl.BACK);

   this.vbo.draw(activeShader, gl.TRIANGLES);

   gl.disable(gl.CULL_FACE);
   gl.disable(gl.BLEND);
};

var RenderUtilities = {
   'init': function(tgl) {
      var gl = tgl.gl;

      // private methods
      var createQuadVbo = function()
      {
         var vbo = new tgl.VertexBuffer();
         vbo.addAttribute("a_Position", gl.FLOAT, 2);
         vbo.addAttribute("a_TexCoord", gl.FLOAT, 2);
         vbo.init(4);
         vbo.copyAttribute('a_Position', new Float32Array([
            -1, -1,
            1, -1,
            -1,  1,
            1,  1,
         ]));
         vbo.copyAttribute('a_TexCoord', new Float32Array([
            0, 0,
            1, 0,
            0, 1,
            1, 1,
         ]));
         return vbo;
      }

      // public methods
      var quadVbo = null;
      RenderUtilities.getQuadVbo = function()
      {
         if(quadVbo == null)
         {
            quadVbo = createQuadVbo();
         }
         return quadVbo;
      };

      var floatToRgbaShader = new tgl.Shader('blit-vert.shader', 'blit-encode-float-rgba-frag.shader');
      RenderUtilities.getFloatToRgbaShader = function()
      {
         return floatToRgbaShader;
      };
   }
};

(function (exports) {
   var formatFloat = function(num)
   {
      // toFixed(): show the first 2 non-zero digits
      var formatted = num.toFixed(1 - Math.floor(Math.log(num)/Math.log(10)));
      return formatted;
   };

   exports.formatFloat = formatFloat;
})(window.Utilities = window.Utilities || {});


var PhotonTracer = function (sigmaS, sigmaA, maxVertexCount) {
   this.rng = new Xorshift32;
   this.sigmaS = sigmaS;
   this.sigmaA = sigmaA;
   this.sigmaT = sigmaA.add(sigmaS);
   this.albedo = sigmaS.divide(this.sigmaT);
   this.sampleSigmaT =
      Math.min(this.sigmaT.x, Math.min(this.sigmaT.y, this.sigmaT.z));
   this.bounce = new Int32Array(maxVertexCount);
   this.pdf = new Float32Array(maxVertexCount);
   this.posX = new Float32Array(maxVertexCount);
   this.posY = new Float32Array(maxVertexCount);
   this.posZ = new Float32Array(maxVertexCount);
   this.colorR = new Float32Array(maxVertexCount);
   this.colorG = new Float32Array(maxVertexCount);
   this.colorB = new Float32Array(maxVertexCount);
};
PhotonTracer.prototype.reset = function () {
   this.rng.reset()
};

//// CanonicalPhotonTracer: class for tracing photons (first pass of photon mapping)
//// The photon will scatter two times. The travel direction will change by 90 degrees
//// in each scatter. (First travel in x+ dir, then in y+ dir, finally in z+ dir.)
var CanonicalPhotonTracer = function(sigmaS, sigmaA, maxVertexCount)
{
   PhotonTracer.call(this, sigmaS, sigmaA, maxVertexCount);
};
// inheritance: CanonicalPhotonTracer -> PhotonTracer
CanonicalPhotonTracer.prototype = Object.create(PhotonTracer.prototype);
CanonicalPhotonTracer.prototype.constructor = CanonicalPhotonTracer;
CanonicalPhotonTracer.prototype.generateSamples =
   function (pathLength, maxVertices, maxSamples) {
      var rng = this.rng;
      var addPhoton = function (i, pos, length) {
         if (i >= maxVertices) return;
         this.pdf[i] = 1.0;
         this.posX[i] = pos.x;
         this.posY[i] = pos.y;
         this.posZ[i] = pos.z;
         this.bounce[i] = length;
      }.bind(this);

      var numSamples = 0;
      var numVertices = 0;
      var length = 0;
      var pos = new Vector(0, 0, 0);
      var dir = new Vector(0, 0, 0);

      while (numVertices < maxVertices && numSamples < maxSamples) {
         if (length == 0) {
            pos.init(0, -2, 0);
            dir.init(1, 0, 0);
            addPhoton(numVertices++, pos, length);
         }
         else if (length == 1) {
            dir.init(0, 1, 0);
         }
         else if (length == 2) {
            dir.init(0, 0, 1);
         }
         else {
            throw new Error(`Unexpected path length: ${length}`);
         }

         var t = -Math.log(1 - rng.next()) / this.sampleSigmaT;
         pos.x += dir.x * t;
         pos.y += dir.y * t;
         pos.z += dir.z * t;
         length++;
         addPhoton(numVertices, pos, length);

         numVertices++;
         if(length == pathLength) {
            numSamples++;
            length = 0;
         }
      }

      return [numSamples, numVertices];
   };


var SearchLightPhotonTracer = function(sigmaS, sigmaA, phaseG, maxVertexCount)
{
   PhotonTracer.call(this, sigmaS, sigmaA, maxVertexCount);
   this.pathDistance = new Float32Array(maxVertexCount);
   this.phaseG = Math.max(-.99, Math.min(.99, phaseG));
   this.INV_FOUR_PI = 1.0 / (4.0 * Math.PI);
};
// inheritance: SearchLightPhotonTracer -> PhotonTracer
SearchLightPhotonTracer.prototype = Object.create(PhotonTracer.prototype);
SearchLightPhotonTracer.prototype.constructor = SearchLightPhotonTracer;

SearchLightPhotonTracer.prototype.evalPhase = function(wi, wo)
{
   var _g = this.phaseG;
   var cosTheta = wi.dot(wo);
   var term = 1.0 + _g*_g - 2.0*_g*cosTheta;
   return this.INV_FOUR_PI * (1.0 - _g * _g) / (term * Math.sqrt(term));
};

SearchLightPhotonTracer.prototype.generateSamples =
   function (power, pathLength, maxVertices, maxSamples, isStartDirFixed = true) {
      var rng = this.rng;
      var phaseG = this.phaseG;

      function sampleSphereUniform() {
         var phi = rng.next() * Math.PI * 2;
         var xi1 = rng.next();
         var sin_theta = 2.0 * Math.sqrt(xi1 - xi1 * xi1);
         var x = Math.cos(phi) * sin_theta;
         var y = Math.sin(phi) * sin_theta;
         var z = 1.0 - 2.0 * xi1;
         return new Vector(x, y, z);
      }

      function samplePhase(dir) {
         var phi = rng.next() * Math.PI * 2;
         var cosTheta;
         if (phaseG == 0) cosTheta = rng.next() * 2 - 1;
         else {
            var tau = (1 - phaseG * phaseG) / (1 + phaseG * (rng.next() * 2 - 1));
            cosTheta = (1 + phaseG * phaseG - tau * tau) / (2 * phaseG)
         }
         var sinTheta = Math.sqrt(Math.max(1 - cosTheta * cosTheta, 0));
         var dx = Math.cos(phi) * sinTheta;
         var dy = Math.sin(phi) * sinTheta;
         var dz = cosTheta;
         if (phaseG != 0) {
            var nx = dir.x,
                 ny = dir.y,
               nz = dir.z;
            var bx, by, bz;
            if (Math.abs(nx) > .99) {
               var l = Math.sqrt(nz * nz + nx * nx);
               bx = -nz / l;
               by = 0;
               bz = nx / l
            } else {
               var l = Math.sqrt(nz * nz + ny * ny);
               bx = 0;
               by = nz / l;
               bz = -ny / l
            }
            var tx = by * nz - bz * ny;
            var ty = bz * nx - bx * nz;
            var tz = bx * ny - by * nx;
            var l = Math.sqrt(tx * tx + ty * ty + tz * tz);
            tx /= l;
            ty /= l;
            tz /= l;
            dir.init(tx * dx + bx * dy + nx * dz, ty * dx + by * dy + ny * dz, tz * dx + bz * dy + nz * dz)
         } else dir.init(dx, dy, dz)
      }
      var addPhoton = function (i, pdf, pos, r, g, b, length, pathDist) {
         if (i >= maxVertices) return;
         this.pdf[i] = pdf;
         this.posX[i] = pos.x;
         this.posY[i] = pos.y;
         this.posZ[i] = pos.z;
         this.colorR[i] = r;
         this.colorG[i] = g;
         this.colorB[i] = b;
         this.bounce[i] = length;
         this.pathDistance[i] = pathDist;
      }.bind(this);
      var numSamples = 0;
      var numVertices = 0;
      var length = 0;
      var r, g, b;
      var pos = new Vector(0, 0, 0);
      var dir = new Vector(0, 0, 0);
      var pathDist = 0;
      while (numVertices < maxVertices && numSamples < maxSamples) {
         if (length == 0) {
            r = power;
            g = power;
            b = power;
            pos.init(0, .75, 0);
            if (isStartDirFixed)
            {
               dir.init(0, -1, 0);
            }
            else
            {
               dir = sampleSphereUniform();
            }
            pathDist = 0;
            addPhoton(numVertices++, 1, pos, r, g, b, length, pathDist);
         }
         var t = -Math.log(1 - rng.next()) / this.sampleSigmaT;
         pos.x += dir.x * t;
         pos.y += dir.y * t;
         pos.z += dir.z * t;
         var disp = new Vector(dir.x * t, dir.y * t, dir.z * t);
         pathDist += disp.length();
         var pdf =
            this.sampleSigmaT * Math.exp(-t * this.sampleSigmaT);
         r *= Math.exp(-t * this.sigmaT.x) * this.sigmaS.x / pdf;
         g *= Math.exp(-t * this.sigmaT.y) * this.sigmaS.y / pdf;
         b *= Math.exp(-t * this.sigmaT.z) * this.sigmaS.z / pdf;
         length++;
         addPhoton(numVertices, pdf, pos, r, g, b, length, pathDist);
         samplePhase(dir);
         numVertices++;
         if (length == pathLength) {
            numSamples++;
            length = 0
         }
      }
      return [numSamples, numVertices]
   };

var app = new TransientPhoton();