#include "simple-common.shader"

// ellipsoid primitive only support transient mode
#define ENABLE_TRANSIENT 1

in vec2 a_Position;      // positions of a quad in [-1, 1]^2
in vec3 a_FocalPos;
in float a_MinTau;
in float a_MaxTau;
in vec3 a_Color;

uniform vec3 u_FrustumCornerBL; // bottom-left
uniform vec3 u_FrustumCornerBR; // bottom-right
uniform vec3 u_FrustumCornerTL; // top-left
uniform vec3 u_FrustumCornerTR; // top-right

out vec3 v_RayDir;
out vec3 v_FocalPos;
out float v_MinTau;
out float v_MaxTau;
out vec3 v_Color;

void main()
{
    gl_Position = vec4(a_Position, 0.0, 1.0);
    vec2 uv = (a_Position + vec2(1.0)) / 2.0;

    float u = uv.x;
    float v = uv.y;
    vec3 rayDir = (1.0 - u) * (1.0 - v) * u_FrustumCornerBL
                + u * (1.0 - v) * u_FrustumCornerBR
                + v * (1.0 - u) * u_FrustumCornerTL
                + u * v * u_FrustumCornerTR
                ;
    v_RayDir = normalize(rayDir);
    v_FocalPos = a_FocalPos;
    v_MinTau = a_MinTau;
    v_MaxTau = a_MaxTau;
    v_Color = a_Color;
}