#include "simple-common.shader"
#include "photon-common.shader"

varying vec4 v_Position;
varying vec4 v_ViewPosition;
varying vec3 v_Color;
varying vec3 v_ViewPhotonDirection;
varying vec3 v_ViewPlaneNormal;

#ifdef ENABLE_TRANSIENT
varying vec2 v_PlanePos;
varying float v_InitDist;

// u_DistInterval.x: minimum light ray path length for path to fall within time interval
// u_DistInterval.y: maximum light ray ...
uniform vec2 u_DistInterval;
#endif
// u_EvalCamTransmittance: if the shader need to calculate camera tranmittance and
//                         multiply it to color
uniform bool u_EvalCamTransmittance;
uniform vec3 u_SigmaT;
uniform bool u_IsPerspective;

// u_CameraWarped: if set to true, take camera time into account when rendering
//                 transient state
uniform bool u_CameraWarped;

void main()
{
    vec3 color = v_Color;

    float camDist = 0.0;
    if (u_EvalCamTransmittance)
    {
        float t = u_IsPerspective ? length(v_ViewPosition.xyz) : abs(v_ViewPosition.z);
        if (u_CameraWarped)
        {
            camDist = t;
        }
        vec3 transmittance = exp(-u_SigmaT*t);
        color *= transmittance;
    }

    if (u_IsPerspective)
    {
        // calculate phase funciton of w_c and w_b
        vec3 w_c = -normalize(v_ViewPosition.xyz);
        vec3 w_b = normalize(v_ViewPhotonDirection);
        // NOTE: assume g = 0.0
        float phase = evalPhase(0.0, w_c, w_b);
        color *= vec3(phase);

        // calculate jacobian term
        float jacobian = abs(dot(-w_c, v_ViewPlaneNormal));
        if (jacobian < 0.01)
        {
            color = vec3(0.0);
        }
        else
        {
            color /= jacobian;
        }
    }

#ifdef ENABLE_TRANSIENT
    float dist = v_InitDist + v_PlanePos.x + v_PlanePos.y + camDist;

    float distFactor = 0.0;
    if(dist > u_DistInterval.x && dist < u_DistInterval.y)
    {
        distFactor = 1.0;
    }
    color *= distFactor;
#endif

    gl_FragColor = vec4(color, 1.0);
}