#include "simple-common.shader"

// atrribute will receive data from a buffer
attribute vec3 a_Position;
attribute mat3 a_Transform;
attribute vec3 a_Offset;

attribute vec3 a_Color;
attribute vec2 a_PlanePos;
attribute float a_InitDist;
attribute vec3 a_PhotonDirection;
attribute vec3 a_PlaneNormal;

uniform mat4 u_ModelViewProjection;
uniform mat4 u_ModelView;

varying vec4 v_Position;
varying vec4 v_ViewPosition;
varying vec3 v_Color;
varying vec3 v_ViewPhotonDirection;
varying vec3 v_ViewPlaneNormal;
#ifdef ENABLE_TRANSIENT
varying vec2 v_PlanePos;
varying float v_InitDist;
#endif

void main()
{
    vec3 worldPos = a_Transform * a_Position + a_Offset;
    gl_Position = u_ModelViewProjection * vec4(worldPos, 1.0);
    v_Position = vec4(worldPos, 1.0);
    v_ViewPosition = u_ModelView * vec4(worldPos, 1.0);
    v_Color = a_Color;
    v_ViewPhotonDirection = (u_ModelView * vec4(a_PhotonDirection, 0.0)).xyz;
    v_ViewPlaneNormal = (u_ModelView * vec4(a_PlaneNormal, 0.0)).xyz;

#ifdef ENABLE_TRANSIENT
    // a_Position is normalized in [0, 1]^2
    v_PlanePos = a_PlanePos * a_Position.xy;
    v_InitDist = a_InitDist;
#endif
}