#include "simple-common.shader"
#include "photon-common.shader"

in vec3 v_ViewPosition;
in vec3 v_ViewSphereCenter;
in vec3 v_Color;
in vec3 v_InDirection;
in float v_Radius;

out vec4 f_Color;

uniform vec2 u_Resolution;
uniform mat4 u_InvProjection;
uniform vec3 u_SigmaT;

bool raySphereIntersect(vec3 o, vec3 d, vec3 c, float r, out float t1, out float t2)
{
    vec3 l = c - o;
    float s = dot(l, d);
    float lsqr = dot(l, l);
    float rsqr = r * r;
    if (s < 0.0 && lsqr > rsqr)
    {
        return false;
    }

    float msqr = lsqr - s * s;
    if (msqr > rsqr)
    {
        return false;
    }

    float q = sqrt(rsqr - msqr);
    t1 = s - q;
    t2 = s + q;
    return true;
}

void main()
{
    // assume orthographic camera
    vec2 uv = gl_FragCoord.xy / u_Resolution.xy;
    vec3 ndc = vec3(uv * 2.0 - 1.0, -1.0);
    vec3 camRayOrigin = (u_InvProjection * vec4(ndc, 1.0)).xyz;
    vec3 camRayDir = vec3(0.0, 0.0, -1.0);

    f_Color = vec4(0.0,0.0,0.0,1.0);
    float t1 = 0.0;
    float t2 = 0.0;
    if (!raySphereIntersect(camRayOrigin, camRayDir, v_ViewSphereCenter, v_Radius, t1, t2))
    {
        return;
    }

    vec2 ts = vec2(t1, t2);
    for (int i = 0; i < 2; ++i)
    {
        float t = ts[i];
        if (t < 0.0)
        {
            continue;
        }

        vec3 hitPoint = camRayOrigin + camRayDir * t;
        vec3 offset = hitPoint - v_ViewSphereCenter;
        vec3 w1 = normalize(offset);
        float t1 = v_Radius;

        if (dot(w1, w1) == 0.f || t1 == 0.f)
        {
            continue;
        }

        vec3 n = t1 * t1 * w1;
        float jacobian = abs(dot(n, camRayDir));
        // if (jacobian < 1e-5f)
        // {
        //     continue;
        // }
        float invJacobian = 1.0 / jacobian;

        vec3 camTrans = exp(-u_SigmaT * t);

        // NOTE: assume g == 0.0
        float lastScatter = evalPhase(0.0, v_InDirection, w1);
        float fp1 = evalPhase(0.0, w1, -camRayDir);
        vec3 estimate = lastScatter * fp1 * u_SigmaT * v_Color * invJacobian * camTrans;

        // NOTE: WARNING: code still produces nan/inf
        // TODO: figure out why
        if (any(isnan(estimate)) || any(isinf(estimate)))
        {
            continue;
        }

        f_Color.xyz += estimate;
    }
}