// reference: photon-planes-min.js on https://benedikt-bitterli.me/photon-planes/webgl/
// Original copyright from Benedikt Bitterli's photon planes WebGL demo.
/*
   Photon planes WebGL demo
   Copyright (c) 2017 Benedikt Bitterli

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/
var tgl = {
    'init': function(gl)
    {
        tgl.gl = gl;
        // var ext = gl.getExtension('ANGLE_instanced_arrays');
        // if (!ext) {
        //     return alert('need ANGLE_instanced_arrays');
        // }
        // tgl.ext = ext;

        function glTypeSize(type) {
            switch (type) {
                case gl.BYTE:
                case gl.UNSIGNED_BYTE:
                    return 1;
                case gl.SHORT:
                case gl.UNSIGNED_SHORT:
                    return 2;
                case gl.INT:
                case gl.UNSIGNED_INT:
                case gl.FLOAT:
                    return 4;
                default:
                    return 0;
            }
        }

        tgl.Texture = function (width, height, channels, isFloat, isLinear = false, isClamped = false, texels = null) {
           var coordMode = isClamped ? gl.CLAMP_TO_EDGE : gl.REPEAT;
           this.type = isFloat ? gl.FLOAT : gl.UNSIGNED_BYTE;

           const formats = [gl.RED, gl.RG, gl.RGB, gl.RGBA];
           const bInternalFormats = [gl.R8, gl.RG8, gl.RGB8, gl.RGBA8];
           const fInternalFormats = [gl.R32F, gl.RB32F, gl.RGB32F, gl.RGBA32F];

           this.format = formats[channels - 1];
           this.internalFormat = isFloat ? fInternalFormats[channels - 1] : bInternalFormats[channels - 1];
           this.width = width;
           this.height = height;
           this.glName = gl.createTexture();
           gl.bindTexture(gl.TEXTURE_2D, this.glName);
           gl.texImage2D(gl.TEXTURE_2D, 0, this.internalFormat, this.width, this.height, 0, this.format, this.type, texels);
           gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, coordMode);
           gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, coordMode);
           this.setSmooth(isLinear);
           this.boundUnit = -1;
        };
        tgl.Texture.prototype.setSmooth = function (smooth) {
           var interpMode = smooth ? gl.LINEAR : gl.NEAREST;
           gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, interpMode);
           gl.texParameteri(gl.TEXTURE_2D,
              gl.TEXTURE_MAG_FILTER, interpMode);
        };
        tgl.Texture.prototype.bind = function (unit) {
           gl.activeTexture(gl.TEXTURE0 + unit);
           gl.bindTexture(gl.TEXTURE_2D, this.glName);
           this.boundUnit = unit;
        };
        tgl.Texture.prototype.toArray = function(brightness = 1.0) {
            // NOTE: only supports float texture for now

            // NOTE: WebGL 1.0 only mandates gl.readPixels() to support the RGBA8 format.
            //       To read a GPU float texture, we need to pack each float value to unsigned byte components.
            if (this.type != gl.FLOAT)
            {
               throw new Error("toArray(): unsupported type: " + this.type);
            }
            // Create RGBA8 texture
            var uintTex = new tgl.Texture(this.width, this.height, 4, false);
            // Create RenderTarget
            var rt = new tgl.RenderTarget(uintTex);
            rt.bind();
            // Render to the RenderTarget, using Float2Rgba shader
            // NOTE: WARNING: will only encode R channel for now
            var shader = RenderUtilities.getFloatToRgbaShader();
            var quadVbo = RenderUtilities.getQuadVbo();

            shader.bind();
            this.bind(0);
            shader.uniformTexture("u_Texture", this);
            shader.uniform1F("u_Factor", brightness);

            quadVbo.bind();
            quadVbo.draw(shader, gl.TRIANGLE_STRIP);

            var uintArray = new Uint8Array(this.width * this.height * 4);
            gl.readPixels(0, 0, this.width, this.height, gl.RGBA, gl.UNSIGNED_BYTE, uintArray);

            // TODO: Destroying webgl objects?

            // Type casting: uint8 * 4 -> float32
            var floatArray = new Float32Array(uintArray.buffer);

            // Debug code: averaging pixels
            // var sum = 0.0;
            // floatArray.forEach(function(elem){
            //    sum += elem;
            // });
            // console.log(sum / (this.width * this.height));

            return floatArray;
        };
        tgl.RenderTarget = function (texture) {
           this.glName = gl.createFramebuffer();
           if(texture != undefined)
           {
              this.bind();
              this.attachTexture(texture);
           }
        };
        tgl.RenderTarget.bound = null;
        tgl.RenderTarget.prototype.bind = function () {
           if(tgl.RenderTarget.bound != this)
           {
              gl.bindFramebuffer(gl.FRAMEBUFFER, this.glName);
              tgl.RenderTarget.bound = this;
           }
        };
        tgl.RenderTarget.unbind = function() {
            if(tgl.RenderTarget.bound != null)
            {
               gl.bindFramebuffer(gl.FRAMEBUFFER, null);
               tgl.RenderTarget.bound = null;
            }
        };
        tgl.RenderTarget.prototype.attachTexture = function (texture, index = 0) {
           this.bind();
           gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0 + index, gl.TEXTURE_2D, texture.glName, 0);
        };
        tgl.RenderTarget.prototype.detachTexture = function (index) {
           this.bind();
           gl.framebufferTexture2D(gl.FRAMEBUFFER, gl.COLOR_ATTACHMENT0 + index, gl.TEXTURE_2D, null, 0);
        };
        tgl.RenderTarget.prototype.clearColor = function (r, g, b, a) {
           this.bind();
           gl.clearColor(r, g, b, a);
           gl.clear(gl.COLOR_BUFFER_BIT);
        };
        // `macros` is a dict containing all preprocessor macros you want to add before the source shader
        // For example, if macros['ENABLE_TRANSIENT'] = 1,
        // the constructor will add "#define ENABLE_TRANSIENT 1" at the beginning of all shader sources.
        tgl.Shader = function(vertName, fragName, macros = undefined, useGLES30 = false)
        {
            var vertSrc = this.resolveShader(vertName);
            var fragSrc = this.resolveShader(fragName);

            if(macros !== undefined)
            {
               var vertDef = "";
               var fragDef = "";
               for(var key in macros)
               {
                  var value = macros[key];
                  var defineStr = `#define ${key} ${value}\n`;
                  vertDef = vertDef.concat(defineStr);
                  fragDef = fragDef.concat(defineStr);
               }
               vertSrc = vertDef + vertSrc;
               fragSrc = fragDef + fragSrc;
            }

            if(useGLES30)
            {
                const GLES30_HEADER = `#version 300 es\n
                #define USE_GLES30 1
                `;
                vertSrc = GLES30_HEADER + vertSrc;
                fragSrc = GLES30_HEADER + fragSrc;
            }

            var vert = this.createShaderObject(gl.VERTEX_SHADER, vertSrc);
            var frag = this.createShaderObject(gl.FRAGMENT_SHADER, fragSrc);

            // reference: https://webglfundamentals.org/webgl/lessons/webgl-fundamentals.html
            var program = gl.createProgram();
            gl.attachShader(program, vert);
            gl.attachShader(program, frag);
            gl.linkProgram(program);
            var success = gl.getProgramParameter(program, gl.LINK_STATUS);
            if(success)
            {
                this.program = program;
            }
            else
            {
                console.log(gl.getProgramInfoLog(program));
                gl.deleteProgram(program);
                this.program = null;
            }
        };
        tgl.Shader.bound = null;

        // resolveShader: returns a shader string with '#include' flattened
        tgl.Shader.prototype.resolveShader = function(fileName, resolvedNames)
        {
            resolvedNames = resolvedNames || new Set();
            // reference: https://benedikt-bitterli.me/photon-planes/webgl/
            var src = ShaderLoader.getShader(fileName);
            // var pattern = new RegExp('#include "(.+)"');

            var pattern = /#include "(.+)"/g;
            var match;

            var matches = [];
            while (match = pattern.exec(src))
            {
               var matchInfo = {
                  'start'   : match.index,
                  'length'  : match[0].length,
                  'fileName': match[1],
               }

               // get the line number of this include
               var srcBefore = src.substring(0, matchInfo.start + matchInfo.length);
               // count number of '\n's
               var lineBreakCount = (srcBefore.match(/[\n]/g) || []).length;
               matchInfo['nextLineNumber'] = lineBreakCount + 2;

               matches.push(matchInfo);
            }

            var resolvedOffset = 0;
            for(var i = 0; i < matches.length; ++i)
            {
               var match = matches[i];
               var resolved = this.resolveShader(match.fileName, resolvedNames);

               if (resolvedNames.has(match.fileName))
               {
                   resolved = "";
               }
               else
               {
                   resolvedNames.add(match.fileName);
               }
               // Use the #line directive to restore the line number to the one
               // before resolving.
               resolved = "#line 0\r\n"
                        + resolved
                        + `\r\n#line ${match.nextLineNumber}`;

               src =  src.slice(0, match.start + resolvedOffset)
                    + resolved
                    + src.slice(match.start + resolvedOffset + match.length);

               resolvedOffset += resolved.length - match.length;
            }
            return src;
        };

        tgl.Shader.prototype.createShaderObject = function(type, source)
        {
            // reference: https://webglfundamentals.org/webgl/lessons/webgl-fundamentals.html
            var shader = gl.createShader(type);
            gl.shaderSource(shader, source);
            gl.compileShader(shader);
            var success = gl.getShaderParameter(shader, gl.COMPILE_STATUS);
            if(success)
            {
                return shader;
            }

            console.log(gl.getShaderInfoLog(shader));
            gl.deleteShader(shader);
        };
        tgl.Shader.prototype.uniformLocation = function(name)
        {
            this.bind();
            var loc = gl.getUniformLocation(this.program, name);
            if(loc == null)
            {
               throw new Error(`cannot get uniform with name \'${name}\'`);
            }
            return loc;
        };
        tgl.Shader.prototype.uniformMatrix = function(name, matrix)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            // gl.uniformMatrix assumes col-major matrix
            gl.uniformMatrix4fv(loc, false, matrix.transpose().m);
        };
        tgl.Shader.prototype.uniformTexture = function(name, texture)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            if(texture.boundUnit < 0)
            {
               throw new Error(`texture bounded unit ${texture.boundUnit} out of range.`);
            }
            gl.uniform1i(loc, texture.boundUnit);
        };
        tgl.Shader.prototype.uniform1F = function(name, f)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            gl.uniform1f(loc, f);
        };
        tgl.Shader.prototype.uniform2F = function(name, f1, f2)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            gl.uniform2f(loc, f1, f2);
        };
        tgl.Shader.prototype.uniform3F = function(name, f1, f2, f3)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            gl.uniform3f(loc, f1, f2, f3);
        };
        tgl.Shader.prototype.uniform1B = function(name, b)
        {
            // Should be able to set bool with gl.uniform1f:
            // ref: https://github.com/KhronosGroup/WebGL/blob/master/sdk/tests/conformance/uniforms/gl-uniform-bool.html
            var f = b ? 1 : 0;
            this.uniform1F(name, f);
        };
        tgl.Shader.prototype.uniform3I = function(name, i1, i2, i3)
        {
            this.bind();
            var loc = this.uniformLocation(name);
            gl.uniform3i(loc, i1, i2, i3);
        };
        tgl.Shader.prototype.bind = function()
        {
            if(tgl.Shader.bound != this)
            {
               gl.useProgram(this.program);
               tgl.Shader.bound = this;
            }
        };
        tgl.VertexBuffer = function()
        {
            this.attributes = [];
            this.elementSize = 0;
            this.isDrawInstanced = false;
        };
        tgl.VertexBuffer.bound = null;
        // sizeX: how many values per row (e.g. for vec3 it's 3)
        // sizeY: how many rows per column (e.g. for mat4x4 it's 4)
        // instanceDivisior: if == 0, the attribute changes for every vertex
        //                   if > 0, the attribute only changes for each `instanceDivisior` vertices
        tgl.VertexBuffer.prototype.addAttribute = function(name, type, sizeX, sizeY = 1, instanceDivisor = 0)
        {
            var size = sizeX * sizeY;
            this.attributes.push({
                'name'  : name,
                'size'  : size,
                'sizeX' : sizeX,
                'sizeY' : sizeY,
                'byteSize' : glTypeSize(type) * size,
                'type'  : type,
                'index' : -1,
                'offsetInBuffer' : -1,
                'sizeInBuffer'   : -1,
                'instanceDivisor': instanceDivisor,
            });
            this.elementSize += glTypeSize(type) * size;

            if (instanceDivisor > 0)
            {
                this.isDrawInstanced = true;
            }
        };
        tgl.VertexBuffer.prototype.init = function(numVertices, numInstances = 0)
        {
            if (this.isDrawInstanced && numVertices != 0 && numInstances == 0)
            {
                // throw new Error("init: you have specified an instanced attribute but numInstances is not specified or is zero.")
            }

            this.buffer = this.buffer || gl.createBuffer();
            this.numVertices = numVertices;
            this.numInstances = numInstances;

            // set the offsetInBuffer and sizeInBuffer field for each attribute
            var offsetInBuffer = 0;
            for(var i = 0; i < this.attributes.length; ++i)
            {
               var attr = this.attributes[i];
               var byteSize = attr.byteSize;
               var sizeInBuffer = attr.instanceDivisor > 0 ?
                      byteSize * numInstances / attr.instanceDivisor
                    : byteSize * numVertices;

               attr.offsetInBuffer = offsetInBuffer;
               attr.sizeInBuffer = sizeInBuffer;

               offsetInBuffer += sizeInBuffer;
            }

            this.bufferSize = offsetInBuffer;

            gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
            gl.bufferData(gl.ARRAY_BUFFER, this.bufferSize, gl.DYNAMIC_DRAW);
        };
        tgl.VertexBuffer.prototype.copyAttribute = function(attrName, arrayBuffer)
        {
            var offsetInBuffer = -1;
            var sizeInBuffer = -1;
            // TODO: optimize whole array search?
            for(var i = 0; i < this.attributes.length; ++i)
            {
               var attr = this.attributes[i];
               if(attr.name == attrName)
               {
                  offsetInBuffer = attr.offsetInBuffer;
                  sizeInBuffer = attr.sizeInBuffer;
                  break;
               }
            }

            gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
            if (arrayBuffer.byteLength != sizeInBuffer) {
               throw new Error("copyAttribute: mismatch between arrayBuffer length and attribute length.");
            }
            gl.bufferSubData(gl.ARRAY_BUFFER, offsetInBuffer, new Float32Array(arrayBuffer, 0, sizeInBuffer / 4));
        };
        tgl.VertexBuffer.prototype.draw = function(shader, primitiveType)
        {
            this.bind();

            for(var i = 0; i < 16; ++i)
            {
                gl.disableVertexAttribArray(i);
            }

            for(var i = 0; i < this.attributes.length; ++i)
            {
                var attr = this.attributes[i];
                var indexBeg = gl.getAttribLocation(shader.program, attr.name);
                attr.index = indexBeg;
                if(indexBeg >= 0)
                {
                    var bytesPerAttr = attr.size * glTypeSize(attr.type);
                    var bytesPerRow = attr.sizeX * glTypeSize(attr.type);
                    for(var row = 0; row < attr.sizeY; ++row)
                    {
                        var index = indexBeg + row;
                        var rowOffset = bytesPerRow * row;
                        // init a row of attr
                        gl.enableVertexAttribArray(index);
                        gl.vertexAttribPointer(index, attr.sizeX, attr.type, false, bytesPerAttr, attr.offsetInBuffer + rowOffset);

                        // instanced draw
                        if (attr.instanceDivisor > 0)
                        {
                            gl.vertexAttribDivisor(index, attr.instanceDivisor);
                        }
                        else
                        {
                            gl.vertexAttribDivisor(index, 0);
                        }
                    }
                }
            }

            if (!this.isDrawInstanced)
            {
                gl.drawArrays(primitiveType, 0, this.numVertices);
            }
            else
            {
                gl.drawArraysInstanced(primitiveType, 0, this.numVertices, this.numInstances);
            }

            for(var i = 0; i < this.attributes.length; ++i)
            {
                var attr = this.attributes[i];
                if(attr.index >= 0)
                {
                    gl.disableVertexAttribArray(attr.index);
                    gl.vertexAttribDivisor(attr.index, 0);
                    attr.index = -1;
                }
            }
        };
        tgl.VertexBuffer.prototype.bind = function()
        {
           if(tgl.VertexBuffer.bound != this)
           {
               gl.bindBuffer(gl.ARRAY_BUFFER, this.buffer);
               tgl.VertexBuffer.bound = this;
           }
        };
    }
}

// simplistic javascript singleton, reference:
// https://www.sitepoint.com/javascript-design-patterns-singleton/
var Shaders = Shaders || {};
var ShaderLoader = (function(){

   function loadShader(fileName)
   {
      // reference: https://stackoverflow.com/questions/5878703/webgl-is-there-an-alternative-to-embedding-shaders-in-html
      var shader;
      var succeed = false;

      $.ajax({
         url: "shaders/" + fileName,
         dataType: 'text',
         async: false,
         success: function(result) {
            shader = result;
            succeed = true;
         }
      });

      if(succeed)
      {
         Shaders[fileName] = shader;
      }

      return succeed;
   }

   return {
      getShader: function(fileName)
      {
         if(!(fileName in Shaders))
         {
            console.warn(`Cannot find shader ${fileName} from shaders.js. Loading from server. Run pack-shaders.py to pack it in!`);
            var succeed = loadShader(fileName);
            if(!succeed)
            {
               throw new Error("cannot load shader: " + fileName);
            }
         }
         var shader = Shaders[fileName];
         // console.log(`Successfully loaded shader ${fileName}.`);
         return shader;
      }
   };
})();
