// reference: photon-planes-min.js on https://benedikt-bitterli.me/photon-planes/webgl/
// Original copyright from Benedikt Bitterli's photon planes WebGL demo.
/*
   Photon planes WebGL demo
   Copyright (c) 2017 Benedikt Bitterli

   Permission is hereby granted, free of charge, to any person obtaining a copy
   of this software and associated documentation files (the "Software"), to deal
   in the Software without restriction, including without limitation the rights
   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
   copies of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in all
   copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
   SOFTWARE.
*/
(function (exports) {
   function stripClass(node, className) {
      node.className = node.className.replace(new RegExp("(?:^|\\s)" + className + "(?!\\S)"), "")
   }

   function addClass(node, className) {
      if (node.className.indexOf(className) == -1) node.className += " " + className
   }
   exports.ButtonGroup = function (targetId, vertical, labels, selectionCallback) {
      this.selectionCallback = selectionCallback;
      this.selectedButton = 0;
      var target = document.getElementById(targetId);
      if (!target) return;
      this.group = document.createElement(vertical ? "ul" : "div");
      this.group.className =
         vertical ? "button-group-vert" : "button-group-horz";
      this.buttons = [];
      for (var i = 0; i < labels.length; ++i) {
         var button = document.createElement(vertical ? "li" : "div");
         button.className = vertical ? "button-vert" : "button-horz";
         button.appendChild(document.createTextNode(labels[i]));
         this.buttons.push(button);
         this.group.appendChild(button);
         button.addEventListener("click", function (idx, event) {
            this.select(idx)
         }.bind(this, i))
      }
      this.select(0);
      target.parentNode.replaceChild(this.group, target);
      this.parent = this.group.parentNode;
   };
   exports.ButtonGroup.prototype.select =
      function (idx) {
         if (idx < 0 || idx >= this.buttons.length) return;
         stripClass(this.buttons[this.selectedButton], "active");
         addClass(this.buttons[idx], "active");
         if (this.selectedButton != idx && this.selectionCallback) this.selectionCallback(idx);
         this.selectedButton = idx
      };
   exports.ProgressBar = function (targetId, hasLabel) {
      var target = document.getElementById(targetId);
      if (!target) return;
      var progressBackground = document.createElement("div");
      progressBackground.className = "progress";
      this.progressBar = document.createElement("div");
      this.progressBar.className = "progress-bar";
      progressBackground.appendChild(this.progressBar);
      this.setProgress(0);
      var parent = target.parentNode;
      parent.replaceChild(progressBackground, target);
      if (hasLabel) {
         this.label = document.createElement("p");
         this.label.className = "progress-label";
         parent.insertBefore(this.label, progressBackground.nextSibling)
      }
      this.parent = parent;
   };
   exports.ProgressBar.prototype.getProgress = function () {
      return this.progressFraction
   };
   exports.ProgressBar.prototype.setProgress = function (progressFraction) {
      this.progressFraction =
         progressFraction;
      this.progressPercentage = Math.min(Math.max(Math.floor(progressFraction * 100), 0), 100);
      this.progressBar.style.width = this.progressPercentage.toString() + "%"
   };
   exports.ProgressBar.prototype.setProgressWithoutTransition = function (progressFraction) {
      addClass(this.progressBar, "notransition");
      this.setProgress(progressFraction);
      this.progressBar.offsetHeight;
      stripClass(this.progressBar, "notransition")
   };
   exports.ProgressBar.prototype.setLabel = function (text) {
      if (this.label) this.label.textContent = text
   };
   exports.ProgressBar.prototype.getProgressPercentage = function () {
      return this.progressPercentage
   };
   var MouseListener = function (target, moveCallback) {
      this.isTouch = false;
      var mouseUpListener = function (event) {
         document.removeEventListener(this.isTouch ? "touchmove" : "mousemove", this.moveListener);
         document.removeEventListener(this.isTouch ? "touchend" : "mouseup", mouseUpListener)
      }.bind(this);

      function mouseDownListener(touch, event) {
         if (touch) this.isTouch = true;
         event.preventDefault();
         this.moveListener = moveCallback.bind(null,
            this.isTouch);
         document.addEventListener(this.isTouch ? "touchmove" : "mousemove", this.moveListener);
         document.addEventListener(this.isTouch ? "touchend" : "mouseup", mouseUpListener)
      }
      target.addEventListener("touchstart", mouseDownListener.bind(this, true));
      target.addEventListener("mousedown", mouseDownListener.bind(this, false))
   };
   exports.Slider = function (targetId, minValue, maxValue, hasLabel, requireInt, callback) {
      var target = document.getElementById(targetId);
      if (!target) return;
      this.sliderBackground = document.createElement("div");
      this.sliderBackground.className = "slider";
      this.minValue = minValue;
      this.maxValue = maxValue;
      this.requireInt = requireInt;
      this.callback = callback;
      this.sliderBar = document.createElement("div");
      this.sliderBar.className = "slider-bar";
      this.sliderBackground.appendChild(this.sliderBar);
      this.sliderHandle = document.createElement("a");
      this.sliderHandle.className = "slider-handle";
      this.sliderBackground.appendChild(this.sliderHandle);
      var that = this;
      new MouseListener(this.sliderHandle, function (isTouch, event) {
         that.mouseMove(isTouch, event)
      });
      var parent =
         target.parentNode;
      this.parent = parent;
      parent.replaceChild(this.sliderBackground, target);
      if (hasLabel) {
         this.label = document.createElement("p");
         this.label.className = "slider-label";
         parent.insertBefore(this.label, this.sliderBackground.nextSibling)
      }
      this.setPosition(.45)
   };
   exports.Slider.prototype.mouseMove = function (isTouch, event) {
      var base = isTouch ? event.touches[0] : event;
      var rect = this.sliderBackground.getBoundingClientRect();
      this.setPosition((base.clientX - rect.left) / (rect.right - rect.left))
   };
   exports.Slider.prototype.setLabel = function (text) {
      if (this.label) this.label.textContent =
         text
   };
   exports.Slider.prototype.getRawPercentage = function (value) {
      var rawPercentage = 100 * (value - this.minValue) / (this.maxValue - this.minValue);
      return rawPercentage;
   };
   exports.Slider.prototype.getPercentage = function (value) {
      var rawPercentage = this.getRawPercentage(value);
      if(this.requireInt)
      {
         rawPercentage = Math.floor(rawPercentage);
      }
      var percentage = Math.max(Math.min(rawPercentage, 100), 0);
      return percentage;
   };
   exports.Slider.prototype.setValue = function (value) {
      value = Math.min(this.maxValue, Math.max(this.minValue, value));
      if (value != this.value) {
         this.value = value;
         var percentage = this.getPercentage(value);
         this.sliderHandle.style.left = this.sliderBar.style.width = percentage.toString() + "%";
         if (this.callback) this.callback(value)
      }
   };
   exports.Slider.prototype.setPosition = function (position) {
      var value = (this.minValue + position * (this.maxValue -
         this.minValue));
      if(this.requireInt)
      {
         value = Math.floor(value);
      }
      this.setValue(value);
   };
   exports.Slider.prototype.show = function (show) {
      var display = show ? "block" : "none";
      this.sliderBackground.style.display = display;
      if (this.label) this.label.style.display = display;
   };
   // Exponential Slider
   exports.ExpSlider = function (targetId, minValue, maxValue, hasLabel, requireInt, callback) {
      exports.Slider.call(this, targetId, minValue, maxValue, hasLabel, requireInt, callback);
      this.logMinValue = Math.log(minValue);
      this.logMaxValue = Math.log(maxValue);
   };
   // inheritance: ExpSlider -> Slider
   exports.ExpSlider.prototype = Object.create(exports.Slider.prototype);
   exports.ExpSlider.prototype.constructor = exports.ExpSlider;

   exports.ExpSlider.prototype.setPosition = function (position) {
      var logValue = (this.logMinValue + position * (this.logMaxValue -
         this.logMinValue));
      var value = Math.exp(logValue);
      if(this.requireInt)
      {
         value = Math.floor(value);
      }
      this.setValue(value);
   };
   exports.ExpSlider.prototype.getRawPercentage = function (value) {
      var logValue = Math.log(value);
      var rawPercentage = 100 * (logValue - this.logMinValue) / (this.logMaxValue - this.logMinValue);
      return rawPercentage;
   };

   exports.ButtonGrid = function (targetId, numCols, labels, selectionCallback) {
      var target = document.getElementById(targetId);
      if (!target) return;
      this.cols = numCols;
      this.selectionCallback = selectionCallback;
      this.selectedButton = 0;
      this.container = document.createElement("div");
      this.container.className = "button-grid";
      this.columns = [];
      for (var i = 0; i < this.cols; ++i) {
         var column = document.createElement("div");
         column.className = "button-grid-column";
         this.container.appendChild(column);
         this.columns.push(column)
      }
      this.cells = [];
      for (var i = 0; i < labels.length; ++i) {
         var column = i % this.cols;
         var cell = document.createElement("div");
         cell.className = "button stretch-button button-grid-button";
         cell.appendChild(document.createTextNode(labels[i]));
         if (i == 0) addClass(cell, "button-grid-tl");
         if (i == this.cols - 1) addClass(cell, "button-grid-tr");
         if (i + this.cols >= labels.length) {
            if (column ==
               0) addClass(cell, "button-grid-bl");
            if (column == this.cols - 1 || i == labels.length - 1) addClass(cell, "button-grid-br")
         }
         cell.addEventListener("click", function (idx, event) {
            this.select(idx)
         }.bind(this, i));
         this.columns[column].appendChild(cell);
         this.cells.push(cell)
      }
      this.select(0);
      this.parent = target.parentNode;
      target.parentNode.replaceChild(this.container, target)
   };
   exports.ButtonGrid.prototype.select = function (idx) {
      if (idx < 0 || idx >= this.cells.length) return;
      stripClass(this.cells[this.selectedButton], "active");
      addClass(this.cells[idx], "active");
      if (this.selectedButton !=
         idx && this.selectionCallback) this.selectionCallback(idx);
      this.selectedButton = idx
   };
   exports.ButtonGrid.prototype.show = function (show) {
      this.container.style.display = show ? "flex" : "none"
   };
   exports.MouseListener = function (target, callback) {
      this.target = target;
      this.callback = callback;
      this.mouseUpHandler = this.mouseUp.bind(this);
      this.mouseMoveHandler = this.mouseMove.bind(this);
      target.addEventListener("mousedown", this.mouseDown.bind(this))
   };
   exports.MouseListener.prototype.mouseDown = function (evt) {
      evt.preventDefault();
      this.mouseStart = this.mapMouseEvent(evt);
      this.callback(this.mouseStart, this.mouseStart);
      document.addEventListener("mouseup", this.mouseUpHandler);
      document.addEventListener("mousemove", this.mouseMoveHandler)
   };
   exports.MouseListener.prototype.mouseUp = function (evt) {
      document.removeEventListener("mouseup", this.mouseUpHandler);
      document.removeEventListener("mousemove", this.mouseMoveHandler)
   };
   exports.MouseListener.prototype.mouseMove = function (evt) {
      this.callback(this.mouseStart, this.mapMouseEvent(evt))
   };
   exports.MouseListener.prototype.mapMouseEvent =
      function (evt) {
         var rect = this.target.getBoundingClientRect();
         return [evt.clientX - rect.left, evt.clientY - rect.top]
      }
})(window.tui = window.tui || {});
